# Copyright (C) 2023 Callum Dickinson
#
# Buildarr is free software: you can redistribute it and/or modify it under the terms of the
# GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# Buildarr is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Buildarr.
# If not, see <https://www.gnu.org/licenses/>.


"""
Sonarr plugin download client settings.
"""


from __future__ import annotations

from logging import getLogger
from typing import Dict, List, Mapping, Union

from buildarr.config import RemoteMapEntry
from pydantic import Field
from typing_extensions import Annotated, Self

from ...api import api_get, api_put
from ...secrets import SonarrSecrets
from ..types import SonarrConfigBase
from .download_clients import (
    DOWNLOADCLIENT_TYPE_MAP,
    Aria2DownloadClient,
    DelugeDownloadClient,
    DownloadstationTorrentDownloadClient,
    DownloadstationUsenetDownloadClient,
    FloodDownloadClient,
    HadoukenDownloadClient,
    NzbgetDownloadClient,
    NzbvortexDownloadClient,
    PneumaticDownloadClient,
    QbittorrentDownloadClient,
    RtorrentDownloadClient,
    SabnzbdDownloadClient,
    TorrentBlackholeDownloadClient,
    TransmissionDownloadClient,
    UsenetBlackholeDownloadClient,
    UtorrentDownloadClient,
    VuzeDownloadClient,
)
from .remote_path_mappings import SonarrRemotePathMappingsSettingsConfig

logger = getLogger(__name__)

DownloadClientType = Union[
    DownloadstationUsenetDownloadClient,
    NzbgetDownloadClient,
    NzbvortexDownloadClient,
    PneumaticDownloadClient,
    SabnzbdDownloadClient,
    UsenetBlackholeDownloadClient,
    Aria2DownloadClient,
    DelugeDownloadClient,
    DownloadstationTorrentDownloadClient,
    FloodDownloadClient,
    HadoukenDownloadClient,
    QbittorrentDownloadClient,
    RtorrentDownloadClient,
    TorrentBlackholeDownloadClient,
    TransmissionDownloadClient,
    UtorrentDownloadClient,
    VuzeDownloadClient,
]


class SonarrDownloadClientsSettingsConfig(SonarrConfigBase):
    """
    Download clients retrieve media files being tracked by Sonarr,
    and store them in a location Sonarr can access to manage the
    downloaded files.

    Download clients that use Usenet or BitTorrent can be configured,
    as well as remote path mappings and other related options.

    ```yaml
    ---

    sonarr:
      settings:
        download_clients:
          enable_completed_download_handling: true
          redownload_failed: true
          delete_unmanaged: true
          definitions:
            Transmission:
              type: "transmission"
              host: "transmission"
              port: 9091
            ...
          remote_path_mappings:
            definitions:
              - host: "transmission"
                remote_path: "/remote/path"
                local_path: "/local/path"
              ...
    ```
    """

    enable_completed_download_handling: bool = True
    """
    Automatically import completed downloads from download clients.
    """

    redownload_failed: bool = True
    """
    Automatically search for and attempt to download a different release.
    """

    delete_unmanaged: bool = False
    """
    Automatically delete download clients not defined in Buildarr.
    """

    definitions: Dict[str, Annotated[DownloadClientType, Field(discriminator="type")]] = {}
    """
    Download client definitions, for connecting with external media downloaders.
    """

    remote_path_mappings = SonarrRemotePathMappingsSettingsConfig()
    """
    Configuration for mapping paths on download client hosts to their counterparts
    on this Sonarr instance.

    For more information, refer to "Configuring remote path mappings".
    """

    _remote_map: List[RemoteMapEntry] = [
        ("enable_completed_download_handling", "enableCompletedDownloadHandling", {}),
        ("redownload_failed", "autoRedownloadFailed", {}),
    ]

    @classmethod
    def from_remote(cls, secrets: SonarrSecrets) -> Self:
        downloadclient_config = api_get(secrets, "/api/v3/config/downloadclient")
        downloadclients = api_get(secrets, "/api/v3/downloadclient")
        tag_ids: Dict[str, int] = (
            {tag["label"]: tag["id"] for tag in api_get(secrets, "/api/v3/tag")}
            if any(downloadclient["tags"] for downloadclient in downloadclients)
            else {}
        )
        return cls(
            **cls.get_local_attrs(cls._remote_map, downloadclient_config),
            definitions={
                dc["name"]: DOWNLOADCLIENT_TYPE_MAP[dc["implementation"]]._from_remote(
                    tag_ids=tag_ids,
                    remote_attrs=dc,
                )
                for dc in downloadclients
            },
            remote_path_mappings=SonarrRemotePathMappingsSettingsConfig._from_remote(
                secrets=secrets,
            ),
        )

    def update_remote(
        self,
        tree: str,
        secrets: SonarrSecrets,
        remote: Self,
        check_unmanaged: bool = False,
    ) -> bool:
        # Update download client-related configuration options.
        config_updated, config_remote_attrs = self.get_update_remote_attrs(
            tree,
            remote,
            self._remote_map,
            check_unmanaged=check_unmanaged,
        )
        if config_updated:
            remote_config = api_get(secrets, "/api/v3/config/downloadclient")
            api_put(
                secrets,
                f"/api/v3/config/downloadclient/{remote_config['id']}",
                {
                    "id": remote_config["id"],
                    "downloadClientWorkingFolders": remote_config["downloadClientWorkingFolders"],
                    **config_remote_attrs,
                },
            )
        # Update download clients.
        definitions_updated = self._update_remote_definitions(
            tree=f"{tree}.definitions",
            secrets=secrets,
            local=self.definitions,
            remote=remote.definitions,
            check_unmanaged=check_unmanaged,
        )
        # Update remote path mappings.
        rpms_updated = self.remote_path_mappings._update_remote(
            tree=f"{tree}.remote_path_mappings",
            secrets=secrets,
            remote=remote.remote_path_mappings,
        )
        #
        return any([config_updated, definitions_updated, rpms_updated])

    def _update_remote_definitions(
        self,
        tree: str,
        secrets: SonarrSecrets,
        local: Mapping[str, DownloadClientType],
        remote: Mapping[str, DownloadClientType],
        check_unmanaged: bool,
    ) -> bool:
        changed = False
        downloadclient_ids: Dict[str, int] = {
            downloadclient_json["name"]: downloadclient_json["id"]
            for downloadclient_json in api_get(secrets, "/api/v3/downloadclient")
        }
        tag_ids: Dict[str, int] = (
            {tag["label"]: tag["id"] for tag in api_get(secrets, "/api/v3/tag")}
            if any(downloadclient.tags for downloadclient in local.values())
            or any(downloadclient.tags for downloadclient in remote.values())
            else {}
        )
        # Create download clients that don't exist yet on the remote,
        # and update in-place ones that do.
        for downloadclient_name, downloadclient in local.items():
            downloadclient_tree = f"{tree}[{repr(downloadclient_name)}]"
            if downloadclient_name not in remote:
                downloadclient._create_remote(
                    tree=downloadclient_tree,
                    secrets=secrets,
                    tag_ids=tag_ids,
                    downloadclient_name=downloadclient_name,
                )
                changed = True
            elif downloadclient._update_remote(
                tree=downloadclient_tree,
                secrets=secrets,
                remote=remote[downloadclient_name],  # type: ignore[arg-type]
                tag_ids=tag_ids,
                downloadclient_id=downloadclient_ids[downloadclient_name],
                downloadclient_name=downloadclient_name,
            ):
                changed = True
        # If `delete_unmanaged` is `True`, remove any download clients on the remote
        # that aren't managed by Buildarr.
        # Otherwise, just log them.
        for downloadclient_name, downloadclient in remote.items():
            if downloadclient_name not in local:
                downloadclient_tree = f"{tree}[{repr(downloadclient_name)}]"
                if self.delete_unmanaged:
                    downloadclient._delete_remote(
                        tree=downloadclient_tree,
                        secrets=secrets,
                        downloadclient_id=downloadclient_ids[downloadclient_name],
                    )
                    changed = True
                else:
                    logger.debug("%s: (...) (unmanaged)", downloadclient_tree)
        # Return the resource changed status.
        return changed



# Copyright (C) 2023 Callum Dickinson
#
# Buildarr is free software: you can redistribute it and/or modify it under the terms of the
# GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# Buildarr is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Buildarr.
# If not, see <https://www.gnu.org/licenses/>.


"""
Sonarr plugin general settings configuration.
"""


from __future__ import annotations

from ipaddress import IPv4Address
from typing import Any, Dict, List, Literal, Mapping, Optional, Set, Tuple, Union

from buildarr.config import RemoteMapEntry
from buildarr.types import BaseEnum, NonEmptyStr, Password, Port
from pydantic import Field, validator
from typing_extensions import Self

from ..api import api_get, api_put
from ..secrets import SonarrSecrets
from .types import SonarrConfigBase


class AuthenticationMethod(BaseEnum):
    """
    Sonarr authentication method.
    """

    none = "none"
    basic = "basic"
    form = "forms"


class CertificateValidation(BaseEnum):
    """
    External site HTTPS certification validation method.
    """

    enabled = "enabled"
    local_disabled = "disabledForLocalAddresses"
    disabled = "disabled"


class ProxyType(BaseEnum):
    """
    Proxy server type.
    """

    http = "http"
    socks4 = "socks4"
    socks5 = "socks5"


class SonarrLogLevel(BaseEnum):
    """
    Log level of the Sonarr application.
    """

    INFO = "info"
    DEBUG = "debug"
    TRACE = "trace"


class UpdateMechanism(BaseEnum):
    """
    Sonarr updating mechanism.
    """

    builtin = "builtIn"
    script = "script"
    external = "external"
    apt = "apt"
    docker = "docker"


class GeneralSettings(SonarrConfigBase):
    """
    Sonarr general settings base class.
    """

    _remote_map: List[RemoteMapEntry]

    @classmethod
    def _from_remote(cls, remote_attrs: Mapping[str, Any]) -> Self:
        return cls(**cls.get_local_attrs(cls._remote_map, remote_attrs))

    def _update_remote_attrs(
        self,
        tree: str,
        secrets: SonarrSecrets,
        remote: Self,
        check_unmanaged: bool = False,
    ) -> Tuple[bool, Dict[str, Any]]:
        return self.get_update_remote_attrs(
            tree,
            remote,
            self._remote_map,
            check_unmanaged=check_unmanaged,
            set_unchanged=True,
        )


class HostGeneralSettings(GeneralSettings):
    """
    Sonarr instance connection and name configuration.

    Many of these settings configure Sonarr's external connection interface.
    If they are changed, the [settings Buildarr uses to connect](host.md) with this
    Sonarr instance may need to be updated, so take care when modifying them.

    **Changing any of these settings require a restart of Sonarr to take effect.**
    """

    # According to docs, IPv6 not supported at this time.
    bind_address: Union[Literal["*"], IPv4Address] = "*"
    """
    Bind address for Sonarr. Set to an IPv4 address bound to a local interface
    or `*` to bind on all interfaces.

    Unless you run Sonarr directly on a host machine (i.e. not via Docker) and
    want Sonarr to only be available on a specific network or interface,
    this generally should be left untouched.
    """

    port: Port = 8989  # type: ignore[assignment]
    """
    Unencrypted (HTTP) listening port for Sonarr.

    If Sonarr is being run via Docker in the default bridge mode,
    this setting shouldn't be changed.
    Instead, change the external port it is bound to using
    `--publish <port number>:8989`.
    """

    ssl_port: Port = 9898  # type: ignore[assignment]
    """
    Encrypted (HTTPS) listening port for Sonarr.

    If Sonarr is being run via Docker in the default bridge mode,
    this setting shouldn't be changed.
    Instead, change the external port it is bound to using
    `--publish <port number>:9898`.
    """

    use_ssl: bool = False
    """
    Enable the encrypted (HTTPS) listening port in Sonarr.
    As Sonarr only supports self-signed certificates, it is recommended
    to put Sonarr behind a HTTPS-terminating reverse proxy such as Nginx, Caddy or Traefik.
    """

    url_base: Optional[str] = None
    """
    Add a prefix to all Sonarr URLs,
    e.g. `http://localhost:8989/<url_base>/settings/general`.

    Generally used to accommodate reverse proxies where Sonarr
    is assigned to a subfolder, e.g. `https://example.com/sonarr`.
    """

    instance_name: NonEmptyStr = "Sonarr"  # type: ignore[assignment]
    """
    Instance name in the browser tab and in syslog.
    """

    _remote_map: List[RemoteMapEntry] = [
        ("bind_address", "bindAddress", {}),
        ("port", "port", {}),
        ("ssl_port", "sslPort", {}),
        ("use_ssl", "enableSsl", {}),
        ("url_base", "urlBase", {"decoder": lambda v: v or None, "encoder": lambda v: v or ""}),
        ("instance_name", "instanceName", {}),
    ]


class SecurityGeneralSettings(GeneralSettings):
    """
    Sonarr instance security (authentication) settings.
    """

    authentication: AuthenticationMethod = AuthenticationMethod.none
    """
    Authentication method for logging into Sonarr.
    By default, do not require authentication.

    Values:

    * `none` - No authentication
    * `basic` - Authentication using HTTP basic auth (browser popup)
    * `form` - Authentication using a login page

    Requires a restart of Sonarr to take effect.
    """

    username: Optional[str] = None
    """
    Username for the administrator user. Required if authentication is enabled.

    Requires a restart of Sonarr to take effect.
    """

    password: Optional[Password] = None
    """
    Password for the administrator user. Required if authentication is enabled.

    Requires a restart of Sonarr to take effect.
    """

    certificate_validation: CertificateValidation = CertificateValidation.enabled
    """
    Change how strict HTTPS certification validation is.
    Do not change unless you understand the risks.

    Values:

    * `enabled` - Validate HTTPS certificates for all hosts
    * `local-disabled` - Disable HTTPS certificate validation for hosts on the local network
    * `disabled` - Disable HTTPS certificate validation completely
    """

    _remote_map: List[RemoteMapEntry] = [
        ("authentication", "authenticationMethod", {}),
        (
            "username",
            "username",
            {
                # Set to default value (`None`) if not found on remote.
                "optional": True,
                # Due to the validator, gets set to `None` if authentication is disabled
                # on the remote instance.
                "decoder": lambda v: v or None,
                # Sonarr isn't too picky about this, but replicate the behaviour of the UI.
                "encoder": lambda v: v or "",
            },
        ),
        (
            "password",
            "password",
            {
                # Set to default value (`None`) if not found on remote.
                "optional": True,
                # Due to the validator, gets set to `None` if authentication is disabled
                # on the remote instance.
                "decoder": lambda v: v or None,
                # Sonarr isn't too picky about this, but replicate the behaviour of the UI.
                "encoder": lambda v: v.get_secret_value() if v else "",
            },
        ),
        ("certificate_validation", "certificateValidation", {}),
    ]

    @validator("username", "password")
    def required_when_auth_enabled(
        cls,
        value: Optional[str],
        values: Dict[str, Any],
    ) -> Optional[str]:
        """
        Enforce the following constraints on the validated attributes:

        * If `authentication` is `none`, set the attribute value to `None`.
        * If `authentication` is a value other than `none` (i.e. require authentication),
          ensure that the attribute set to a value other than `None`.

        This will apply to both the local Buildarr configuration and
        the remote Sonarr instance configuration.

        Args:
            value (Optional[str]): Value to validate
            values (Dict[str, Any]): Configuration attributes

        Raises:
            ValueError: If the attribute is required but empty

        Returns:
            Validated attribute value
        """
        if values["authentication"] == AuthenticationMethod.none:
            return None
        elif not value:
            raise ValueError("required when 'authentication' is not set to 'none'")
        return value


class ProxyGeneralSettings(GeneralSettings):
    """
    Proxy configuration for Sonarr.
    """

    enable: bool = False
    """
    Use a proxy server to access the Internet.
    """

    proxy_type: ProxyType = ProxyType.http
    """
    Type of proxy to connect to.

    Values:

    * `http` - HTTP(S) proxy
    * `socks4` - SOCKSv4 proxy
    * `socks5` - SOCKSv5 proxy (Tor is supported)
    """

    # TODO: Enforce constraint
    hostname: Optional[str] = None
    """
    Proxy server hostname.

    Required if using a proxy is enabled.
    """

    port: Port = 8080  # type: ignore[assignment]
    """
    Proxy server access port.
    """

    username: Optional[str] = None
    """
    Username to authenticate with.
    Only enter if authentication is required by the proxy.
    """

    password: Optional[Password] = None
    """
    Password for the proxy user.
    Only enter if authentication is required by the proxy.
    """

    ignored_addresses: Set[NonEmptyStr] = set()
    """
    List of domains/addresses which bypass the proxy. Wildcards (`*`) are supported.
    """

    bypass_proxy_for_local_addresses: bool = True
    """
    Do not use the proxy to access local network addresses.
    """

    _remote_map: List[RemoteMapEntry] = [
        ("enable", "proxyEnabled", {}),
        ("proxy_type", "proxyType", {}),
        (
            "hostname",
            "proxyHostname",
            {"decoder": lambda v: v or None, "encoder": lambda v: v or ""},
        ),
        ("port", "proxyPort", {}),
        (
            "username",
            "proxyUsername",
            {"decoder": lambda v: v or None, "encoder": lambda v: v or ""},
        ),
        (
            "password",
            "proxyPassword",
            {
                "decoder": lambda v: v or None,
                "encoder": lambda v: v.get_secret_value() if v else "",
            },
        ),
        (
            "ignored_addresses",
            "proxyBypassFilter",
            {
                "decoder": lambda v: (
                    set(addr.strip() for addr in v.split(",")) if v and v.strip() else []
                ),
                "encoder": lambda v: ",".join(sorted(v)) if v else "",
            },
        ),
        ("bypass_proxy_for_local_addresses", "proxyBypassLocalAddresses", {}),
    ]


class LoggingGeneralSettings(GeneralSettings):
    """
    Logging configuration for the Sonarr application.
    """

    log_level: SonarrLogLevel = SonarrLogLevel.INFO
    """
    Verbosity of logging output.

    Values:

    * `INFO` - Standard log output
    * `DEBUG` - Debugging log output
    * `TRACE` - Trace diagnostics log output
    """

    _remote_map: List[RemoteMapEntry] = [("log_level", "logLevel", {})]


class AnalyticsGeneralSettings(GeneralSettings):
    """
    Configuration of analytics and telemetry from within Sonarr.
    """

    send_anonymous_usage_data: bool = True
    """
    Send anonymous usage and error information to Sonarr's servers.

    This includes information on your browser, which Sonarr Web UI pages you use,
    error reporting and OS/runtime versions. This information is reportedly used
    to prioritise features and bug fixes.

    Requires a restart of Sonarr to take effect.
    """

    _remote_map: List[RemoteMapEntry] = [("send_anonymous_usage_data", "analyticsEnabled", {})]


class UpdatesGeneralSettings(GeneralSettings):
    """
    Settings for updating Sonarr.
    """

    branch: NonEmptyStr = "main"  # type: ignore[assignment]
    """
    Branch used by the external update mechanism.
    Changing this value has no effect on Docker installations.

    If unsure, leave this undefined in Buildarr and use the value already set in Sonarr.
    """

    automatic: bool = False
    """
    Automatically download and install updates.
    Manual updates can still be performed from System -> Updates.

    This option must be left set to `false` on Docker installations.
    """

    # script_path is required when mechanism is "script"
    # script_path should be absolute only
    mechanism: UpdateMechanism = UpdateMechanism.docker
    """
    Set the mechanism for updating Sonarr.
    Must be set to `docker` on Docker installations.

    Values:

    * `builtin` - Sonarr built-in updater mechanism
    * `script` - Use the configured update script
    * `external` - External update mechanism
    * `apt` - Debian APT package
    * `docker` - Docker image
    """

    # TODO: Constraint - required if update mechanism is "script"
    script_path: Optional[str] = None
    """
    Path to a custom script that takes an extracted update package
    and handles the remainder of the update process.

    Required if `mechanism` is set to `script`.
    """

    _remote_map: List[RemoteMapEntry] = [
        ("branch", "branch", {}),
        ("automatic", "updateAutomatically", {}),
        ("mechanism", "updateMechanism", {}),
        (
            "script_path",
            "updateScriptPath",
            {"decoder": lambda v: v or None, "encoder": lambda v: v or ""},
        ),
    ]


class BackupGeneralSettings(GeneralSettings):
    """
    Settings for Sonarr automatic backups.
    """

    folder: NonEmptyStr = "Backups"  # type: ignore[assignment]
    """
    Folder to backup Sonarr data to.

    Relative paths will be under Sonarr's AppData directory.
    """

    interval: int = Field(7, ge=1, le=7)  # days
    """
    Interval between automatic backups, in days.

    Must be set somewhere between 1 and 7 days.
    """

    retention: int = Field(28, ge=1, le=90)  # days
    """
    Retention period for backups, in days.
    Backups older than the retention period will be cleaned up automatically.

    Must be set somewhere between 1 and 90 days.
    """

    _remote_map: List[RemoteMapEntry] = [
        ("folder", "backupFolder", {}),
        ("interval", "backupInterval", {}),
        ("retention", "backupRetention", {}),
    ]


class SonarrGeneralSettingsConfig(SonarrConfigBase):
    """
    Sonarr general settings.
    """

    host: HostGeneralSettings = HostGeneralSettings()
    security: SecurityGeneralSettings = SecurityGeneralSettings()
    proxy: ProxyGeneralSettings = ProxyGeneralSettings()
    logging: LoggingGeneralSettings = LoggingGeneralSettings()
    analytics: AnalyticsGeneralSettings = AnalyticsGeneralSettings()
    updates: UpdatesGeneralSettings = UpdatesGeneralSettings()
    backup: BackupGeneralSettings = BackupGeneralSettings()

    @classmethod
    def from_remote(cls, secrets: SonarrSecrets) -> Self:
        settings = api_get(secrets, "/api/v3/config/host")
        return cls(
            host=HostGeneralSettings._from_remote(settings),
            security=SecurityGeneralSettings._from_remote(settings),
            proxy=ProxyGeneralSettings._from_remote(settings),
            logging=LoggingGeneralSettings._from_remote(settings),
            analytics=AnalyticsGeneralSettings._from_remote(settings),
            updates=UpdatesGeneralSettings._from_remote(settings),
            backup=BackupGeneralSettings._from_remote(settings),
        )

    def update_remote(
        self,
        tree: str,
        secrets: SonarrSecrets,
        remote: Self,
        check_unmanaged: bool = False,
    ) -> bool:
        host_updated, host_attrs = self.host._update_remote_attrs(
            tree=f"{tree}.host",
            secrets=secrets,
            remote=remote.host,
            check_unmanaged=check_unmanaged,
        )
        security_updated, security_attrs = self.security._update_remote_attrs(
            tree=f"{tree}.security",
            secrets=secrets,
            remote=remote.security,
            check_unmanaged=check_unmanaged,
        )
        proxy_updated, proxy_attrs = self.proxy._update_remote_attrs(
            tree=f"{tree}.proxy",
            secrets=secrets,
            remote=remote.proxy,
            check_unmanaged=check_unmanaged,
        )
        logging_updated, logging_attrs = self.logging._update_remote_attrs(
            tree=f"{tree}.logging",
            secrets=secrets,
            remote=remote.logging,
            check_unmanaged=check_unmanaged,
        )
        analytics_updated, analytics_attrs = self.analytics._update_remote_attrs(
            tree=f"{tree}.analytics",
            secrets=secrets,
            remote=remote.analytics,
            check_unmanaged=check_unmanaged,
        )
        updates_updated, updates_attrs = self.updates._update_remote_attrs(
            tree=f"{tree}.updates",
            secrets=secrets,
            remote=remote.updates,
            check_unmanaged=check_unmanaged,
        )
        backup_updated, backup_attrs = self.backup._update_remote_attrs(
            tree=f"{tree}.backup",
            secrets=secrets,
            remote=remote.backup,
            check_unmanaged=check_unmanaged,
        )
        if any(
            [
                host_updated,
                security_updated,
                proxy_updated,
                logging_updated,
                analytics_updated,
                updates_updated,
                backup_updated,
            ],
        ):
            remote_config = api_get(secrets, "/api/v3/config/host")
            api_put(
                secrets,
                f"/api/v3/config/host/{remote_config['id']}",
                {
                    # There are some undocumented values that are not
                    # set by Buildarr. Pass those through unmodified.
                    **remote_config,
                    **host_attrs,
                    **security_attrs,
                    **proxy_attrs,
                    **logging_attrs,
                    **analytics_attrs,
                    **updates_attrs,
                    **backup_attrs,
                },
            )
            return True
        return False



# Copyright (C) 2023 Callum Dickinson
#
# Buildarr is free software: you can redistribute it and/or modify it under the terms of the
# GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# Buildarr is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Buildarr.
# If not, see <https://www.gnu.org/licenses/>.


"""
Sonarr plugin profiles settings configuration.
"""

from __future__ import annotations

from ..types import SonarrConfigBase
from .delay import SonarrDelayProfilesSettingsConfig
from .language import SonarrLanguageProfilesSettingsConfig
from .quality import SonarrQualityProfilesSettingsConfig
from .release import SonarrReleaseProfilesSettingsConfig


class SonarrProfilesSettingsConfig(SonarrConfigBase):
    """
    Sonarr plugin profiles settings configuration.
    """

    quality_profiles: SonarrQualityProfilesSettingsConfig = SonarrQualityProfilesSettingsConfig()
    language_profiles: SonarrLanguageProfilesSettingsConfig = SonarrLanguageProfilesSettingsConfig()
    delay_profiles: SonarrDelayProfilesSettingsConfig = SonarrDelayProfilesSettingsConfig()
    release_profiles: SonarrReleaseProfilesSettingsConfig = SonarrReleaseProfilesSettingsConfig()



# Copyright (C) 2023 Callum Dickinson
#
# Buildarr is free software: you can redistribute it and/or modify it under the terms of the
# GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# Buildarr is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Buildarr.
# If not, see <https://www.gnu.org/licenses/>.


"""
Sonarr plugin quality settings configuration object.
"""


from __future__ import annotations

import json

from pathlib import Path
from typing import Any, Dict, Mapping, Optional, cast

from buildarr.config import ConfigBase
from buildarr.config.exceptions import ConfigTrashIDNotFoundError
from buildarr.types import TrashID
from pydantic import Field, validator
from typing_extensions import Self

from ..api import api_get, api_put
from ..secrets import SonarrSecrets
from .types import SonarrConfigBase

QUALITYDEFINITION_MAX = 400
"""
The upper bound for the maximum quality allowed in a quality definition.
"""


class QualityDefinition(SonarrConfigBase):
    """
    Manually set quality definitions can have the following parameters.
    """

    title: Optional[str] = None
    """
    The name of the quality in the GUI.

    If unset, set to an empty string or `None`, it will always be set to the
    name of the quality itself. (e.g. For the `Bluray-480p` quality, the GUI title
    will also be `Bluray-480p`)
    """

    min: float = Field(..., ge=0, le=QUALITYDEFINITION_MAX - 1)
    """
    The minimum Megabytes per Minute (MB/min) a quality can have.
    Must be set at least 1MB/min lower than `max`.

    The minimum value is `0`, and the maximum value is `399`.
    """

    # Note: No 'pref' field like in Radarr until V4

    max: Optional[float] = Field(..., ge=1, le=QUALITYDEFINITION_MAX)
    """
    The maximum Megabytes per Minute (MB/min) a quality can have.
    Must be set at least 1MB/min higher than `min`.

    If set to `None` or `400`, the maximum bit rate will be unlimited.

    If not set to `None`, the minimum value is `1`, and the maximum value is `400`.
    """

    @validator("max")
    def validate_min_max(
        cls,
        value: Optional[float],
        values: Mapping[str, Any],
    ) -> Optional[float]:
        quality_max = value
        quality_max_val = (
            min(quality_max, QUALITYDEFINITION_MAX)
            if quality_max is not None
            else QUALITYDEFINITION_MAX
        )
        try:
            quality_min: float = values["min"]
            if quality_max_val - quality_min < 1:
                raise ValueError(
                    f"'max' ({quality_max_val}) is not "
                    f"at least 1 greater than 'min' ({quality_min})",
                )
        except KeyError:
            # `min` only doesn't exist when it failed type validation.
            # If it doesn't exist, skip validation that uses it.
            pass
        if quality_max_val >= QUALITYDEFINITION_MAX:
            return None
        return quality_max


class SonarrQualitySettingsConfig(ConfigBase):
    """
    Quality definitions are used to set the permitted bit rates for each quality level.

    These can either be set manually within Buildarr, or pre-made profiles can be
    imported from TRaSH-Guides.

    ```yaml
    sonarr:
      settings:
        quality:
          trash_id: "bef99584217af744e404ed44a33af589" # series
          definitions:
            Bluray-480p: # "Quality" column name (not "Title")
              min: 2
              max: 100
            # Add additional override quality definitions here
    ```

    Quality definition profiles retrieved from TRaSH-Guides are automatically
    kept up to date by Buildarr, with the latest values being pushed to Sonarr
    on an update run.

    For more information, refer to the guides from
    [WikiArr](https://wiki.servarr.com/sonarr/settings#quality-1)
    and [TRaSH-Guides](https://trash-guides.info/Sonarr/Sonarr-Quality-Settings-File-Size/).
    """

    # When defined, all explicitly defined quality definitions override the Trash version.
    trash_id: Optional[TrashID] = None
    """
    Trash ID of the TRaSH-Guides quality definition profile to load default values from.

    If there is an update in the profile, the quality definitions will be updated accordingly.
    """

    definitions: Dict[str, QualityDefinition] = {}
    """
    Explicitly set quality definitions here.

    The key of the definition is the "Quality" column of the Quality Definitions page
    in Sonarr, **not** "Title".

    If `trash_id` is set, any values set here will override the default values provided
    from the TRaSH-Guides quality definition profile.

    If `trash_id` is not set, only explicitly defined quality definitions are managed,
    and quality definitions not set within Buildarr are left unmodified.
    """

    @property
    def uses_trash_metadata(self) -> bool:
        """
        A flag determining whether or not this configuration uses TRaSH-Guides metadata.

        Returns:
            `True` if TRaSH-Guides metadata is used, otherwise `False`
        """
        return bool(self.trash_id)

    def _render_trash_metadata(self, trash_metadata_dir: Path) -> None:
        """
        Render configuration attributes obtained from TRaSH-Guides, in-place.

        Args:
            trash_metadata_dir (Path): TRaSH-Guides metadata directory.
        """
        if not self.trash_id:
            return
        for quality_file in (
            trash_metadata_dir / "docs" / "json" / "sonarr" / "quality-size"
        ).iterdir():
            with quality_file.open() as f:
                quality_json = json.load(f)
                if cast(str, quality_json["trash_id"]).lower() == self.trash_id:
                    for definition_json in quality_json["qualities"]:
                        definition_name = definition_json["quality"]
                        if definition_name not in self.definitions:
                            self.definitions[definition_name] = QualityDefinition(
                                title=None,
                                min=definition_json["min"],
                                max=definition_json["max"],
                            )
                    return
        raise ConfigTrashIDNotFoundError(
            f"Unable to find Sonarr quality definition file with trash ID '{self.trash_id}'",
        )

    @classmethod
    def from_remote(cls, secrets: SonarrSecrets) -> Self:
        return cls(
            definitions={
                definition_json["quality"]["name"]: QualityDefinition(
                    title=(
                        definition_json["title"]
                        if definition_json["title"] != definition_json["quality"]["name"]
                        else None
                    ),
                    min=definition_json["minSize"],
                    max=definition_json.get("maxSize", None),
                )
                for definition_json in api_get(secrets, "/api/v3/qualitydefinition")
            },
        )

    def update_remote(
        self,
        tree: str,
        secrets: SonarrSecrets,
        remote: Self,
        check_unmanaged: bool = False,
    ) -> bool:
        changed = False
        remote_definitions_json = {
            definition_json["id"]: definition_json
            for definition_json in api_get(secrets, "/api/v3/qualitydefinition")
        }
        definition_ids: Dict[str, int] = {
            definition_json["quality"]["name"]: definition_id
            for definition_id, definition_json in remote_definitions_json.items()
        }
        for definition_name, local_definition in self.definitions.items():
            updated, remote_attrs = local_definition.get_update_remote_attrs(
                tree=f"{tree}[{repr(definition_name)}]",
                remote=remote.definitions[definition_name],
                remote_map=[
                    ("title", "title", {"encoder": lambda v: v or definition_name}),
                    ("min", "minSize", {}),
                    ("max", "maxSize", {}),
                ],
            )
            if updated:
                definition_id = definition_ids[definition_name]
                api_put(
                    secrets,
                    f"/api/v3/qualitydefinition/{definition_id}",
                    {**remote_definitions_json[definition_id], **remote_attrs},
                )
                changed = True
        return changed

    # Tell Pydantic to validate in-place assignments of attributes.
    # This ensures that any validators that parse attributes to consistent values run.
    class Config(SonarrConfigBase.Config):
        validate_assignment = True



# Copyright (C) 2023 Callum Dickinson
#
# Buildarr is free software: you can redistribute it and/or modify it under the terms of the
# GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# Buildarr is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Buildarr.
# If not, see <https://www.gnu.org/licenses/>.


"""
Sonarr plugin media management settings configuration.
"""


from __future__ import annotations

from logging import getLogger
from typing import Any, Dict, List, Optional, Set

from buildarr.config import RemoteMapEntry
from buildarr.types import BaseEnum, NonEmptyStr
from pydantic import Field
from typing_extensions import Self

from ..api import api_delete, api_get, api_post, api_put
from ..secrets import SonarrSecrets
from .types import SonarrConfigBase

logger = getLogger(__name__)


class MultiEpisodeStyle(BaseEnum):
    """
    Multi-episode style enumeration.
    """

    extend = 0
    duplicate = 1
    repeat = 2
    scene = 3
    range = 4
    prefixed_range = 5


class EpisodeTitleRequired(BaseEnum):
    """
    Episode title required enumeration.
    """

    always = "always"
    bulk_season_releases = "bulkSeasonReleases"
    never = "never"


class PropersAndRepacks(BaseEnum):
    """
    Propers and repacks configuration enumeration.
    """

    prefer_and_upgrade = "preferAndUpgrade"
    do_not_upgrade_automatically = "doNotUpgrade"
    do_not_prefer = "doNotPrefer"


class RescanSeriesFolderAfterRefresh(BaseEnum):
    """
    Enumeration for rescan series folder after refresh.
    """

    always = "always"
    after_manual_refresh = "afterManual"
    never = "never"


class ChangeFileDate(BaseEnum):
    """
    Change file date settings enumeration.
    """

    none = "none"
    local_air_date = "localAirDate"
    utc_air_date = "utcAirDate"


class ChmodFolder(BaseEnum):
    """
    Read-write permissions for media folders.
    """

    drwxr_xr_x = "755"
    drwxrwxr_x = "775"
    drwxrwx___ = "770"
    drwxr_x___ = "750"
    drwxrwxrwx = "777"

    @classmethod
    def validate(cls, v: Any) -> ChmodFolder:
        """
        Ensure that octal and decimal integers are both read properly by Buildarr.
        """
        try:
            return cls(oct(v)[2:] if isinstance(v, int) else v)
        except ValueError:
            try:
                return cls[v.replace("-", "_")]
            except (TypeError, KeyError):
                raise ValueError(f"Invalid {cls.__name__} name or value: {v}") from None


class SonarrMediaManagementSettingsConfig(SonarrConfigBase):
    """
    Naming, file management and root folder configuration.

    ```yaml
    sonarr:
      settings:
        media_management:
          ...
    ```

    For more information on how to configure these options correctly,
    refer to these guides from
    [WikiArr](https://wiki.servarr.com/sonarr/settings#media-management)
    and [TRaSH-Guides](https://trash-guides.info/Sonarr/Sonarr-recommended-naming-scheme).
    """

    # Episode Naming
    rename_episodes: bool = False
    """
    Rename imported files to the defined standard format.

    Sonarr will use the existing file name if renaming is disabled.
    """

    replace_illegal_characters: bool = True
    """
    Replace illegal characters within the file name.

    If set to `False`, Sonarr will remove them instead.
    """

    standard_episode_format: NonEmptyStr = (
        "{Series TitleYear} - "  # type: ignore[assignment]
        "S{season:00}E{episode:00} - "
        "{Episode CleanTitle} "
        "[{Preferred Words }{Quality Full}]"
        "{[MediaInfo VideoDynamicRangeType]}"
        "{[Mediainfo AudioCodec}{ Mediainfo AudioChannels]}"
        "{MediaInfo AudioLanguages}"
        "{[MediaInfo VideoCodec]}"
        "{-Release Group}"
    )
    """
    File renaming format for a standard episode file.

    The default specified here is the current TRaSH-Guides recommended format,
    but it will not be applied to the Sonarr instance unless it is explicitly
    defined in the configuration file.
    """

    daily_episode_format: NonEmptyStr = (
        "{Series TitleYear} - "  # type: ignore[assignment]
        "{Air-Date} - "
        "{Episode CleanTitle} "
        "[{Preferred Words }{Quality Full}]"
        "{[MediaInfo VideoDynamicRangeType]}"
        "{[Mediainfo AudioCodec}{ Mediainfo AudioChannels]}"
        "{MediaInfo AudioLanguages}"
        "{[MediaInfo VideoCodec]}"
        "{-Release Group}"
    )
    """
    File renaming format for a daily episode file.

    The default specified here is the current TRaSH-Guides recommended format,
    but it will not be applied to the Sonarr instance unless it is explicitly
    defined in the configuration file.
    """

    anime_episode_format: NonEmptyStr = (
        "{Series TitleYear} - "  # type: ignore[assignment]
        "S{season:00}E{episode:00} - "
        "{absolute:000} - "
        "{Episode CleanTitle} "
        "[{Preferred Words }{Quality Full}]"
        "{[MediaInfo VideoDynamicRangeType]}"
        "[{MediaInfo VideoBitDepth}bit]"
        "{[MediaInfo VideoCodec]}"
        "[{Mediainfo AudioCodec} { Mediainfo AudioChannels}]"
        "{MediaInfo AudioLanguages}"
        "{-Release Group}"
    )
    """
    File renaming format for an anime episode file.

    The default specified here is the current TRaSH-Guides recommended format,
    but it will not be applied to the Sonarr instance unless it is explicitly
    defined in the configuration file.
    """

    series_folder_format: NonEmptyStr = "{Series TitleYear}"  # type: ignore[assignment]
    """
    Renaming format for a series folder.

    The default specified here is the current TRaSH-Guides recommended format,
    but it will not be applied to the Sonarr instance unless it is explicitly
    defined in the configuration file.
    """

    season_folder_format: NonEmptyStr = "Season {season:00}"  # type: ignore[assignment]
    """
    Renaming format for a season folder of a series.

    The default specified here is the current TRaSH-Guides recommended format,
    but it will not be applied to the Sonarr instance unless it is explicitly
    defined in the configuration file.
    """

    specials_folder_format: NonEmptyStr = "Specials"  # type: ignore[assignment]
    """
    Renaming format for a specials folder of a series.

    The default specified here is the current TRaSH-Guides recommended format,
    but it will not be applied to the Sonarr instance unless it is explicitly
    defined in the configuration file.
    """

    multiepisode_style: MultiEpisodeStyle = MultiEpisodeStyle.range
    """
    Formatting style for the episode numbers of a multi-episode media file.
    """

    # Folders
    create_empty_series_folders: bool = False
    """
    Create missing series folders during disk scan.
    """

    delete_empty_folders: bool = False
    """
    Delete empty series and season folders during disk scan and when
    episode files are deleted.
    """

    # Importing
    episode_title_required: EpisodeTitleRequired = EpisodeTitleRequired.always
    """
    Prevent importing for up to 48 hours if the episode title
    is in the naming format and the episode title is TBA.

    Values:

    * `always`
    * `bulk-season-releases`
    * `never`

    ```yaml
    sonarr:
      settings:
        media_management:
          episode_title_required: "always"
    ```
    """

    skip_free_space_check: bool = False
    """
    Skip the free space check for the series root folder.

    Only enable when Sonarr is unable to detect free space from your series root folder.
    """

    minimum_free_space: int = Field(100, ge=100)  # MB
    """
    Prevent import if it would leave less than the specified
    amount of disk space (in megabytes) available.

    Minimum value is 100 MB.
    """

    use_hardlinks: bool = True
    """
    Use hard links when trying to copy files from torrents that are still being seeded.

    Occasionally, file locks may prevent renaming files that are being seeded.
    You may temporarily disable seeding and use Sonarr's rename function as a work around.
    """

    import_extra_files: bool = False
    """
    Import matching extra files (subtitles, `.nfo` file, etc) after importing an episode file.
    """

    # File Management
    unmonitor_deleted_episodes: bool = False
    """
    Episodes deleted from disk are automatically unmonitored in Sonarr.
    """

    propers_and_repacks: PropersAndRepacks = PropersAndRepacks.do_not_prefer
    """
    Whether or not to automatically upgrade to Propers/Repacks.

    Values:

    * `prefer-and-upgrade`
    * `do-not-upgrade-automatically`
    * `do-not-prefer`

    ```yaml
    sonarr:
      settings:
        media_management:
          propers_and_repacks: "do-not-prefer"
    ```

    Use 'Do not Prefer' to sort by preferred word score over propers/repacks.
    Use 'Prefer and Upgrade' for automatic upgrades to propers/repacks.
    """

    analyze_video_files: bool = True
    """
    Extract video information such as resolution, runtime and codec information
    from files.

    This requires Sonarr to read parts of the file, which may cause high disk
    or network activity during scans.
    """

    rescan_series_folder_after_refresh: RescanSeriesFolderAfterRefresh = (
        RescanSeriesFolderAfterRefresh.always
    )
    """
    Rescan the series folder after refreshing the series.

    Values:

    * `always`
    * `after_manual_refresh`
    * `never`

    ```yaml
    sonarr:
      settings:
        media_management:
          rescan_series_folder_after_refresh: "always"
    ```

    NOTE: Sonarr will not automatically detect changes to files
    if this option is not set to `always`.
    """

    change_file_date: ChangeFileDate = ChangeFileDate.none
    """
    Change file date on import/rescan.

    Values:

    * `none`
    * `local-air-date`
    * `utc-air-date`

    ```yaml
    sonarr:
      settings:
        media_management:
          change_file_date: "none"
    ```
    """

    recycling_bin: Optional[NonEmptyStr] = None
    """
    Episode files will go here when deleted instead of being permanently deleted.
    """

    recycling_bin_cleanup: int = Field(7, ge=0)  # days
    """
    Files in the recycle bin older than the selected number of days
    will be cleaned up automatically.

    Set to 0 to disable automatic cleanup.
    """

    # Permissions
    set_permissions: bool = False
    """
    Set whether or not `chmod` should run when files are imported/renamed.

    If you're unsure what this and the `chmod`/`chown` series of attributes do,
    do not alter them.
    """

    chmod_folder: ChmodFolder = ChmodFolder.drwxr_xr_x
    """
    Permissions to set on media folders and files during import/rename.
    File permissions are set without execute bits.

    This only works if the user running Sonarr is the owner of the file.
    It's better to ensure the download client sets the permissions properly.

    Values:

    * `drwxr-xr-x`/`755`
    * `drwxrwxr-x`/`775`
    * `drwxrwx---`/`770`
    * `drwxr-x---`/`750`
    * `drwxrwxrwx`/`777`

    ```yaml
    sonarr:
      settings:
        media_management:
          chmod_folder: "drwxr-xr-x"
    ```
    """

    chown_group: Optional[str] = None
    """
    Group name or gid. Use gid for remote file systems.

    This only works if the user running Sonarr is the owner of the file.
    It's better to ensure the download client uses the same group as Sonarr.
    """

    root_folders: Set[NonEmptyStr] = set()
    """
    This allows you to create a root path for a place to either
    place new imported downloads, or to allow Sonarr to scan existing media.

    ```yaml
    sonarr:
      settings:
        media_management:
          root_folders:
            - "/path/to/rootfolder"
    ```
    """

    delete_unmanaged_root_folders: bool = False
    """
    Delete root folder definitions from Sonarr if they are not
    explicitly defined in Buildarr.

    Before enabling this option, ensure all the root folders
    you want Sonarr to scan are defined in Buildarr,
    as Sonarr might remove imported media from its database
    when root folder definitions are deleted.

    *New in version 0.1.2.*
    """

    _naming_remote_map: List[RemoteMapEntry] = [
        # Episode Naming
        ("rename_episodes", "renameEpisodes", {}),
        ("replace_illegal_characters", "replaceIllegalCharacters", {}),
        ("standard_episode_format", "standardEpisodeFormat", {}),
        ("daily_episode_format", "dailyEpisodeFormat", {}),
        ("anime_episode_format", "animeEpisodeFormat", {}),
        ("series_folder_format", "seriesFolderFormat", {}),
        ("season_folder_format", "seasonFolderFormat", {}),
        ("specials_folder_format", "specialsFolderFormat", {}),
        ("multiepisode_style", "multiEpisodeStyle", {}),
    ]
    _mediamanagement_remote_map: List[RemoteMapEntry] = [
        # Folders
        ("create_empty_series_folders", "createEmptySeriesFolders", {}),
        ("delete_empty_folders", "deleteEmptyFolders", {}),
        # Importing
        ("episode_title_required", "episodeTitleRequired", {}),
        ("skip_free_space_check", "skipFreeSpaceCheckWhenImporting", {}),
        ("minimum_free_space", "minimumFreeSpaceWhenImporting", {}),
        ("use_hardlinks", "copyUsingHardlinks", {}),
        ("import_extra_files", "importExtraFiles", {}),
        # File Management
        ("unmonitor_deleted_episodes", "autoUnmonitorPreviouslyDownloadedEpisodes", {}),
        ("propers_and_repacks", "downloadPropersAndRepacks", {}),
        ("analyze_video_files", "enableMediaInfo", {}),
        ("rescan_series_folder_after_refresh", "rescanAfterRefresh", {}),
        ("change_file_date", "fileDate", {}),
        (
            "recycling_bin",
            "recycleBin",
            {"decoder": lambda v: v or None, "encoder": lambda v: v or ""},
        ),
        ("recycling_bin_cleanup", "recycleBinCleanupDays", {}),
        # Permissions
        ("set_permissions", "setPermissionsLinux", {}),
        ("chmod_folder", "chmodFolder", {}),
        (
            "chown_group",
            "chownGroup",
            {"decoder": lambda v: v or None, "encoder": lambda v: v or ""},
        ),
    ]

    @classmethod
    def from_remote(cls, secrets: SonarrSecrets) -> Self:
        return cls(
            # Episode Naming
            **cls.get_local_attrs(
                cls._naming_remote_map,
                api_get(secrets, "/api/v3/config/naming"),
            ),
            # All other sections except Root Folders
            **cls.get_local_attrs(
                cls._mediamanagement_remote_map,
                api_get(secrets, "/api/v3/config/mediamanagement"),
            ),
            # Root Folders
            root_folders=[rf["path"] for rf in api_get(secrets, "/api/v3/rootfolder")],
        )

    def update_remote(
        self,
        tree: str,
        secrets: SonarrSecrets,
        remote: Self,
        check_unmanaged: bool = False,
    ) -> bool:
        return any(
            [
                # Episode Naming
                self._update_remote_naming(
                    tree=tree,
                    secrets=secrets,
                    remote=remote,
                    check_unmanaged=check_unmanaged,
                ),
                # All other sections except Root Folders
                self._update_remote_mediamanagement(
                    tree=tree,
                    secrets=secrets,
                    remote=remote,
                    check_unmanaged=check_unmanaged,
                ),
                # Root Folders
                self._update_remote_rootfolder(
                    tree,
                    secrets=secrets,
                    remote=remote,
                    check_unmanaged=check_unmanaged,
                ),
            ],
        )

    def _update_remote_naming(
        self,
        tree: str,
        secrets: SonarrSecrets,
        remote: Self,
        check_unmanaged: bool = False,
    ) -> bool:
        updated, remote_attrs = self.get_update_remote_attrs(
            tree=tree,
            remote=remote,
            remote_map=self._naming_remote_map,
            check_unmanaged=check_unmanaged,
        )
        if updated:
            api_put(
                secrets,
                f"/api/v3/config/naming/{api_get(secrets, '/api/v3/config/naming')['id']}",
                remote_attrs,
            )
            return True
        return False

    def _update_remote_mediamanagement(
        self,
        tree: str,
        secrets: SonarrSecrets,
        remote: Self,
        check_unmanaged: bool = False,
    ) -> bool:
        updated, remote_attrs = self.get_update_remote_attrs(
            tree,
            remote,
            self._mediamanagement_remote_map,
            check_unmanaged=check_unmanaged,
            set_unchanged=True,
        )
        if updated:
            config_id = api_get(secrets, "/api/v3/config/mediamanagement")["id"]
            api_put(
                secrets,
                f"/api/v3/config/mediamanagement/{config_id}",
                {"id": config_id, **remote_attrs},
            )
            return True
        return False

    def _update_remote_rootfolder(
        self,
        tree: str,
        secrets: SonarrSecrets,
        remote: Self,
        check_unmanaged: bool = False,
    ) -> bool:
        tree = f"{tree}.root_folders"
        changed = False
        current_root_folders: Dict[str, int] = {
            rf["path"]: rf["id"] for rf in api_get(secrets, "/api/v3/rootfolder")
        }
        expected_root_folders = set(self.root_folders)
        if self.delete_unmanaged_root_folders:
            i = -1
            for root_folder, root_folder_id in current_root_folders.items():
                if root_folder not in expected_root_folders:
                    logger.info("%s[%i]: %s -> (deleted)", tree, i, repr(str(root_folder)))
                    api_delete(secrets, f"/api/v3/rootfolder/{root_folder_id}")
                    changed = True
                    i -= 1
        for i, root_folder in enumerate(self.root_folders):
            if root_folder in current_root_folders:
                logger.debug("%s[%i]: %s (exists)", tree, i, repr(str(root_folder)))
            else:
                logger.info("%s[%i]: %s -> (created)", tree, i, repr(str(root_folder)))
                api_post(secrets, "/api/v3/rootfolder", {"path": str(root_folder)})
                changed = True
        return changed



# Copyright (C) 2023 Callum Dickinson
#
# Buildarr is free software: you can redistribute it and/or modify it under the terms of the
# GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# Buildarr is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Buildarr.
# If not, see <https://www.gnu.org/licenses/>.


"""
Sonarr plugin metadata settings configuration.
"""


from __future__ import annotations

from typing import Any, Dict, List, Mapping, Optional, Tuple, Type

from buildarr.config import ConfigBase, RemoteMapEntry
from typing_extensions import Self

from ..api import api_get, api_put
from ..secrets import SonarrSecrets
from .types import SonarrConfigBase


class Metadata(SonarrConfigBase):
    """
    Metadata definition base class.
    """

    enable: bool = False
    """
    When set to `True`, enables creating metadata files in the given format.
    """

    _implementation_name: str
    _implementation: str
    _config_contract: str
    _base_remote_map: List[RemoteMapEntry] = [("enable", "enable", {})]
    _remote_map: List[RemoteMapEntry]

    @classmethod
    def _from_remote(cls, metadata: Dict[str, Any]) -> Self:
        return cls(**cls.get_local_attrs(cls._base_remote_map + cls._remote_map, metadata))

    def _update_remote(
        self,
        tree: str,
        secrets: SonarrSecrets,
        remote: Self,
        metadata: Mapping[str, Any],
        check_unmanaged: bool = False,
    ) -> bool:
        updated, remote_attrs = self.get_update_remote_attrs(
            tree,
            remote,
            self._base_remote_map + self._remote_map,
            check_unmanaged=check_unmanaged,
            set_unchanged=True,
        )
        if updated:
            api_put(
                secrets,
                f"/api/v3/metadata/{metadata['id']}",
                {
                    "id": metadata["id"],
                    "name": self._implementation_name,
                    "implementationName": self._implementation_name,
                    "implementation": self._implementation,
                    "configContract": self._config_contract,
                    **remote_attrs,
                },
            )
            return True
        return False


class KodiEmbyMetadata(Metadata):
    """
    Output metadata files in a format suitable for Kodi (XBMC) or Emby.

    ```yaml
    sonarr:
      settings:
        metadata:
          kodi_emby:
            enable: true
            series_metadata: true
            series_metadata_url: true
            episode_metadata: true
            series_images: true
            season_images: true
            episode_images: true
    ```
    """

    series_metadata: bool = False
    """
    Create `tvshow.nfo` with full series metadata.
    """

    series_metadata_url: bool = False
    """
    Add the TVDB show URL to `tvshow.nfo`. Can be combined with `series_metadata`.
    """

    episode_metadata: bool = False
    """
    Create episode-specific metadata as `<filename>.nfo`.
    """

    series_images: bool = False
    """
    Save series images to `fanart.jpg`, `poster.jpg` and `banner.jpg`.
    """

    season_images: bool = False
    """
    Save season images to `season##-poster.jpg`/`season-specials-poster.jpg`
    and `season##-banner.jpg`/`season-specials-banner.jpg`.
    """

    episode_images: bool = False
    """
    Save episode images to `<filename>-thumb.jpg`.
    """

    _implementation_name: str = "Kodi (XBMC) / Emby"
    _implementation: str = "XbmcMetadata"
    _config_contract: str = "XbmcMetadataSettings"
    _remote_map: List[RemoteMapEntry] = [
        ("series_metadata", "seriesMetadata", {"is_field": True}),
        ("series_metadata_url", "seriesMetadataUrl", {"is_field": True}),
        ("episode_metadata", "episodeMetadata", {"is_field": True}),
        ("series_images", "seriesImages", {"is_field": True}),
        ("season_images", "seasonImages", {"is_field": True}),
        ("episode_images", "episodeImages", {"is_field": True}),
    ]


class RoksboxMetadata(Metadata):
    """
    Output metadata files in a format suitable for Roksbox.

    ```yaml
    sonarr:
      settings:
        metadata:
          roksbox:
            enable: true
            episode_metadata: true
            series_images: true
            season_images: true
            episode_images: true
    ```
    """

    episode_metadata: bool = False
    """
    Create episode-specific metadata as `Season##/<filename>.xml`.
    """

    series_images: bool = False
    """
    Save series images to `<Series Title>.jpg`.
    """

    season_images: bool = False
    """
    Save season images to `Season ##.jpg`.
    """

    episode_images: bool = False
    """
    Save episode images to `Season##/<filename>.jpg`.
    """

    _implementation_name: str = "Roksbox"
    _implementation: str = "RoksboxMetadata"
    _config_contract: str = "RoksboxMetadataSettings"
    _remote_map: List[RemoteMapEntry] = [
        ("episode_metadata", "episodeMetadata", {"is_field": True}),
        ("series_images", "seriesImages", {"is_field": True}),
        ("season_images", "seasonImages", {"is_field": True}),
        ("episode_images", "episodeImages", {"is_field": True}),
    ]


class WdtvMetadata(Metadata):
    """
    Output metadata files in a format suitable for WDTV.

    ```yaml
    sonarr:
      settings:
        metadata:
          wdtv:
            enable: true
            episode_metadata: true
            series_images: true
            season_images: true
            episode_images: true
    ```
    """

    episode_metadata: bool = False
    """
    Create episode-specific metadata as `<filename>.nfo`.
    """

    series_images: bool = False
    """
    Save series images to `fanart.jpg`, `poster.jpg` and `banner.jpg`.
    """

    season_images: bool = False
    """
    Save as images to `season##-poster.jpg`/`season-specials-poster.jpg`
    and `season##-banner.jpg`/`season-specials-banner.jpg`.
    """

    episode_images: bool = False
    """
    Save episode images to `<filename>-thumb.jpg`.
    """

    _implementation_name: str = "WDTV"
    _implementation: str = "WdtvMetadata"
    _config_contract: str = "WdtvMetadataSettings"
    _remote_map: List[RemoteMapEntry] = [
        ("episode_metadata", "episodeMetadata", {"is_field": True}),
        ("series_images", "seriesImages", {"is_field": True}),
        ("season_images", "seasonImages", {"is_field": True}),
        ("episode_images", "episodeImages", {"is_field": True}),
    ]


METADATA_TYPES: Tuple[Type[Metadata], ...] = (KodiEmbyMetadata, RoksboxMetadata, WdtvMetadata)
METADATA_TYPE_MAP: Dict[str, Type[Metadata]] = {
    metadata_type._implementation: metadata_type for metadata_type in METADATA_TYPES
}


class SonarrMetadataSettingsConfig(ConfigBase):
    """
    Sonarr metadata settings.
    Implementation wise each metadata is a unique object, updated using separate requests.
    """

    kodi_emby = KodiEmbyMetadata()
    roksbox = RoksboxMetadata()
    wdtv = WdtvMetadata()

    @classmethod
    def from_remote(cls, secrets: SonarrSecrets) -> Self:
        kodi_emby_metadata: Optional[Dict[str, Any]] = None
        roksbox_metadata: Optional[Dict[str, Any]] = None
        wdtv_metadata: Optional[Dict[str, Any]] = None
        for metadata in api_get(secrets, "/api/v3/metadata"):
            if metadata["implementation"] == KodiEmbyMetadata._implementation:
                kodi_emby_metadata = metadata
            elif metadata["implementation"] == RoksboxMetadata._implementation:
                roksbox_metadata = metadata
            elif metadata["implementation"] == WdtvMetadata._implementation:
                wdtv_metadata = metadata
        if kodi_emby_metadata is None:
            raise RuntimeError(
                "Unable to find Kodi (XBMC)/Emby metadata on Sonarr, database might be corrupt",
            )
        if roksbox_metadata is None:
            raise RuntimeError(
                "Unable to find Roksbox metadata on Sonarr, database might be corrupt",
            )
        if wdtv_metadata is None:
            raise RuntimeError(
                "Unable to find WDTV metadata on Sonarr, database might be corrupt",
            )
        return cls(
            kodi_emby=KodiEmbyMetadata._from_remote(kodi_emby_metadata),
            roksbox=RoksboxMetadata._from_remote(roksbox_metadata),
            wdtv=WdtvMetadata._from_remote(wdtv_metadata),
        )

    def update_remote(
        self,
        tree: str,
        secrets: SonarrSecrets,
        remote: Self,
        check_unmanaged: bool = False,
    ) -> bool:
        kodi_emby_metadata: Optional[Dict[str, Any]] = None
        roksbox_metadata: Optional[Dict[str, Any]] = None
        wdtv_metadata: Optional[Dict[str, Any]] = None
        for metadata in api_get(secrets, "/api/v3/metadata"):
            if metadata["implementation"] == KodiEmbyMetadata._implementation:
                kodi_emby_metadata = metadata
            elif metadata["implementation"] == RoksboxMetadata._implementation:
                roksbox_metadata = metadata
            elif metadata["implementation"] == WdtvMetadata._implementation:
                wdtv_metadata = metadata
        if kodi_emby_metadata is None:
            raise RuntimeError(
                "Unable to find Kodi (XBMC)/Emby metadata on Sonarr, database might be corrupt",
            )
        if roksbox_metadata is None:
            raise RuntimeError(
                "Unable to find Roksbox metadata on Sonarr, database might be corrupt",
            )
        if wdtv_metadata is None:
            raise RuntimeError(
                "Unable to find WDTV metadata on Sonarr, database might be corrupt",
            )
        return any(
            [
                self.kodi_emby._update_remote(
                    tree=f"{tree}.kodi_emby",
                    secrets=secrets,
                    remote=remote.kodi_emby,
                    metadata=kodi_emby_metadata,
                    check_unmanaged=check_unmanaged,
                ),
                self.roksbox._update_remote(
                    tree=f"{tree}.roksbox",
                    secrets=secrets,
                    remote=remote.roksbox,
                    metadata=roksbox_metadata,
                    check_unmanaged=check_unmanaged,
                ),
                self.wdtv._update_remote(
                    tree=f"{tree}.wdtv",
                    secrets=secrets,
                    remote=remote.wdtv,
                    metadata=wdtv_metadata,
                    check_unmanaged=check_unmanaged,
                ),
            ],
        )



# Copyright (C) 2023 Callum Dickinson
#
# Buildarr is free software: you can redistribute it and/or modify it under the terms of the
# GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# Buildarr is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Buildarr.
# If not, see <https://www.gnu.org/licenses/>.


"""
Sonarr plugin language profile configuration.
"""


from __future__ import annotations

from logging import getLogger
from typing import Any, Dict, List, Mapping, Optional, Sequence, Set

from buildarr.config import RemoteMapEntry
from buildarr.types import BaseEnum
from pydantic import Field, validator
from typing_extensions import Annotated, Self

from ...api import api_delete, api_get, api_post, api_put
from ...secrets import SonarrSecrets
from ..types import SonarrConfigBase

logger = getLogger(__name__)


class Language(BaseEnum):
    """
    These are the available languages that can be selected in a language profile.

    * Arabic
    * Bulgarian
    * Chinese
    * Czech
    * Danish
    * Dutch
    * English
    * Finnish
    * Flemish
    * French
    * German
    * Greek
    * Hebrew
    * Hindi
    * Hungarian
    * Icelandic
    * Italian
    * Japanese
    * Korean
    * Lithuanian
    * Malayalam
    * Norwegian
    * Polish
    * Portuguese
    * Russian
    * Spanish
    * Swedish
    * Turkish
    * Ukrainian
    * Vietnamese
    """

    arabic = "Arabic"
    bulgarian = "Bulgarian"
    chinese = "Chinese"
    czech = "Czech"
    danish = "Danish"
    dutch = "Dutch"
    english = "English"
    finnish = "Finnish"
    flemish = "Flemish"
    french = "French"
    german = "German"
    greek = "Greek"
    hebrew = "Hebrew"
    hindi = "Hindi"
    hungarian = "Hungarian"
    icelandic = "Icelandic"
    italian = "Italian"
    japanese = "Japanese"
    korean = "Korean"
    lithuanian = "Lithuanian"
    malayalam = "Malayalam"
    norwegian = "Norwegian"
    polish = "Polish"
    portuguese = "Portuguese"
    russian = "Russian"
    spanish = "Spanish"
    swedish = "Swedish"
    turkish = "Turkish"
    ukrainian = "Ukrainian"
    unknown = "Unknown"
    vietnamese = "Vietnamese"


class LanguageProfile(SonarrConfigBase):
    """
    A language profile is defined under the `language_profiles` block as shown below.

    ```yaml
    ...
      language_profiles:
        definitions:
          Anime: # Name of the language profile
            upgrades_allowed: true
            upgrade_until: "Japanese" # Required if upgrades are allowed
            languages: # Required
              - "Japanese"
              - "English"
    ```
    """

    upgrades_allowed: bool = False
    """
    Enable automatic upgrading if a version of a media file
    in a more preferred language becomes available.

    If disabled, languages will not be upgraded.
    """

    languages: Annotated[List[Language], Field(min_items=1)]
    """
    The languages episodes are allowed to be in.
    The order of the list determines priority (highest priority first, lowest priority last).

    Use the name of the language in English (e.g. `Japanese`, not `Nihongo` or `日本語`).

    ```yaml
    ...
      languages:
        - "Japanese"
        - "English"
    ```

    At least one language must be specified.
    """

    upgrade_until: Optional[Language] = None
    """
    The highest priority language to upgrade an episode to.
    Usually this would be set to the highest priority language in the profile.

    This attribute is required if `upgrades_allowed` is set to `True`.
    """

    @validator("languages")
    def validate_languages(cls, value: List[Language]) -> List[Language]:
        language_set: Set[Language] = set()
        for language in value:
            if language in language_set:
                raise ValueError(f"duplicate entries of language '{language.value}' exist")
            else:
                language_set.add(language)
        return value

    @validator("upgrade_until")
    def validate_upgrade_until(
        cls,
        value: Optional[Language],
        values: Dict[str, Any],
    ) -> Optional[Language]:
        try:
            upgrades_allowed: bool = values["upgrades_allowed"]
            languages: Sequence[Language] = values["languages"]
        except KeyError:
            return value
        # If `upgrades_allowed` is `False`, set `upgrade_until` to `None`
        # to make sure Buildarr ignores whatever it is currently set to
        # on the remote instance.
        if not upgrades_allowed:
            return None
        # Subsequent checks now assume that `upgrades_allowed` is `True`,
        # this parameter is required and defined to a valid value.
        if not value:
            raise ValueError("required if 'upgrades_allowed' is True")
        for language in languages:
            if value == language:
                break
        else:
            raise ValueError("must be set to a value enabled in 'languages'")
        return value

    @classmethod
    def _get_remote_map(
        cls,
        language_ids: Mapping[Language, int],
    ) -> List[RemoteMapEntry]:
        return [
            ("upgrades_allowed", "upgradeAllowed", {}),
            (
                "upgrade_until",
                "cutoff",
                {
                    "decoder": lambda v: Language(v["name"]),
                    "root_encoder": lambda vs: cls._upgrade_until_encoder(
                        language_ids=language_ids,
                        language=vs.upgrade_until if vs.upgrade_until else vs.languages[0],
                    ),
                },
            ),
            (
                "languages",
                "languages",
                {
                    "decoder": lambda v: list(
                        reversed([ln["language"]["name"] for ln in v if ln["allowed"]]),
                    ),
                    "encoder": lambda v: cls._languages_encoder(language_ids, v),
                },
            ),
        ]

    @classmethod
    def _upgrade_until_encoder(
        cls,
        language_ids: Mapping[Language, int],
        language: Language,
    ) -> Dict[str, Any]:
        return {"id": language_ids[language], "name": language.value}

    @classmethod
    def _languages_encoder(
        cls,
        language_ids: Mapping[Language, int],
        languages: List[Language],
    ) -> List[Dict[str, Any]]:
        return list(
            reversed(
                [
                    {
                        "language": {"id": language_ids[ln], "name": ln.value},
                        "allowed": True,
                    }
                    for ln in languages
                ]
                + [
                    {
                        "language": {"id": language_ids[ln], "name": ln.value},
                        "allowed": False,
                    }
                    for ln in (lang for lang in Language if lang not in languages)
                ],
            ),
        )

    @classmethod
    def _from_remote(
        cls,
        language_ids: Mapping[Language, int],
        remote_attrs: Mapping[str, Any],
    ) -> Self:
        return cls(
            **cls.get_local_attrs(cls._get_remote_map(language_ids), remote_attrs),
        )

    def _create_remote(
        self,
        tree: str,
        secrets: SonarrSecrets,
        profile_name: str,
        language_ids: Mapping[Language, int],
    ) -> None:
        api_post(
            secrets,
            "/api/v3/languageprofile",
            {
                "name": profile_name,
                **self.get_create_remote_attrs(tree, self._get_remote_map(language_ids)),
            },
        )

    def _update_remote(
        self,
        tree: str,
        secrets: SonarrSecrets,
        remote: LanguageProfile,
        profile_id: int,
        profile_name: str,
        language_ids: Mapping[Language, int],
    ) -> bool:
        changed, remote_attrs = self.get_update_remote_attrs(
            tree,
            remote,
            self._get_remote_map(language_ids),
            check_unmanaged=True,
            set_unchanged=True,
        )
        if changed:
            api_put(
                secrets,
                f"/api/v3/languageprofile/{profile_id}",
                {"id": profile_id, "name": profile_name, **remote_attrs},
            )
            return True
        return False

    def _delete_remote(self, tree: str, secrets: SonarrSecrets, profile_id: int) -> None:
        logger.info("%s: (...) -> (deleted)", tree)
        api_delete(secrets, f"/api/v3/languageprofile/{profile_id}")


class SonarrLanguageProfilesSettingsConfig(SonarrConfigBase):
    """
    Configuration parameters for controlling how Buildarr handles language profiles.
    """

    delete_unmanaged = False
    """
    Automatically delete language profiles not defined in Buildarr.
    """

    definitions: Dict[str, LanguageProfile] = {}
    """
    Define language profiles to configure on Sonarr here.

    If there are no language profiles defined and `delete_unmanaged` is `False`,
    Buildarr will not modify existing language profiles, but if `delete_unmanaged` is `True`,
    **Buildarr will delete all existing profiles. Be careful when using `delete_unmanaged`.**
    """

    @classmethod
    def from_remote(cls, secrets: SonarrSecrets) -> Self:
        language_ids = {
            Language(language["language"]["name"]): language["language"]["id"]
            for language in api_get(secrets, "/api/v3/languageprofile/schema")["languages"]
        }
        return cls(
            definitions={
                profile["name"]: LanguageProfile._from_remote(language_ids, profile)
                for profile in api_get(secrets, "/api/v3/languageprofile")
            },
        )

    def update_remote(
        self,
        tree: str,
        secrets: SonarrSecrets,
        remote: Self,
        check_unmanaged: bool = False,
    ) -> bool:
        #
        changed = False
        #
        profile_ids: Dict[str, int] = {
            profile_json["name"]: profile_json["id"]
            for profile_json in api_get(secrets, "/api/v3/languageprofile")
        }
        language_ids = {
            Language(language["language"]["name"]): language["language"]["id"]
            for language in api_get(secrets, "/api/v3/languageprofile/schema")["languages"]
        }
        # # Only works on Sonarr V4
        # try:
        #     language_ids: Dict[Language, int] = {
        #         Language(language["name"]): language["id"]
        #         for language in api_get(secrets, "/api/v3/language")
        #     }
        # # Compatible with Sonarr V3, deprecated on Sonarr V4
        # except SonarrAPIError as err:
        #     if err.response.status_code == 404:
        #         ...
        #     else:
        #         raise
        #
        for profile_name, profile in self.definitions.items():
            profile_tree = f"{tree}.definitions[{repr(profile_name)}]"
            #
            if profile_name not in remote.definitions:
                profile._create_remote(
                    tree=profile_tree,
                    secrets=secrets,
                    profile_name=profile_name,
                    language_ids=language_ids,
                )
                changed = True
            #
            elif profile._update_remote(
                tree=profile_tree,
                secrets=secrets,
                remote=remote.definitions[profile_name],
                profile_id=profile_ids[profile_name],
                profile_name=profile_name,
                language_ids=language_ids,
            ):
                changed = True
        #
        for profile_name, profile in remote.definitions.items():
            if profile_name not in self.definitions:
                profile_tree = f"{tree}.definitions[{repr(profile_name)}]"
                if self.delete_unmanaged:
                    profile._delete_remote(
                        tree=profile_tree,
                        secrets=secrets,
                        profile_id=profile_ids[profile_name],
                    )
                    changed = True
                else:
                    logger.debug("%s: (...) (unmanaged)", profile_tree)
        #
        return changed



# Copyright (C) 2023 Callum Dickinson
#
# Buildarr is free software: you can redistribute it and/or modify it under the terms of the
# GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# Buildarr is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Buildarr.
# If not, see <https://www.gnu.org/licenses/>.


"""
Sonarr plugin for Buildarr root module.
"""


from __future__ import annotations

from importlib_metadata import PackageNotFoundError, version as package_version

__all__ = ["__version__"]

try:
    __version__: str = package_version("buildarr-sonarr")
except PackageNotFoundError:
    __version__ = "0.4.0"



# Copyright (C) 2023 Callum Dickinson
#
# Buildarr is free software: you can redistribute it and/or modify it under the terms of the
# GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# Buildarr is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Buildarr.
# If not, see <https://www.gnu.org/licenses/>.


"""
Sonarr plugin tags settings configuration.
"""


from __future__ import annotations

from logging import getLogger
from typing import Dict, List

from buildarr.types import NonEmptyStr
from typing_extensions import Self

from ..api import api_get, api_post
from ..secrets import SonarrSecrets
from .types import SonarrConfigBase

logger = getLogger(__name__)


class SonarrTagsSettingsConfig(SonarrConfigBase):
    """
    Tags are used to associate media files with certain resources (e.g. release profiles).

    ```yaml
    sonarr:
      settings:
        tags:
          definitions:
            - "example1"
            - "example2"
    ```

    To be able to use those tags in Buildarr, they need to be defined
    in this configuration section.
    """

    delete_unused: bool = False
    """
    Delete tags that are not used by any resource in Buildarr.

    Note that tags not being used in Buildarr are not necessarily
    unused by Sonarr, so be careful about when to use this option.

    Sonarr appears to periodically clean up unused tags,
    so in most cases there is no need to enable this option.
    """

    definitions: List[NonEmptyStr] = []
    """
    Define tags that are used within Buildarr here.

    If they are not defined here, you may get errors resulting from non-existent
    tags from either Buildarr or Sonarr.
    """

    @classmethod
    def from_remote(cls, secrets: SonarrSecrets) -> Self:
        return cls(
            definitions=[tag["label"] for tag in api_get(secrets, "/api/v3/tag")],
        )

    def update_remote(
        self,
        tree: str,
        secrets: SonarrSecrets,
        remote: Self,
        check_unmanaged: bool = False,
    ) -> bool:
        # This only does creations and updates.
        # Deletes (and empty tag list prints) are done AFTER all other modifications are made.
        # TODO: Implement tag deletions.
        changed = False
        current_tags: Dict[str, int] = {
            tag["label"]: tag["id"] for tag in api_get(secrets, "/api/v3/tag")
        }
        if self.definitions:
            for i, tag in enumerate(self.definitions):
                if tag in current_tags:
                    logger.debug("%s.definitions[%i]: %s (exists)", tree, i, repr(tag))
                else:
                    logger.info("%s.definitions[%i]: %s -> (created)", tree, i, repr(tag))
                    api_post(secrets, "/api/v3/tag", {"label": tag})
                    changed = True
        return changed



# Copyright (C) 2023 Callum Dickinson
#
# Buildarr is free software: you can redistribute it and/or modify it under the terms of the
# GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# Buildarr is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Buildarr.
# If not, see <https://www.gnu.org/licenses/>.


"""
Sonarr plugin UI settings configuration object.
"""


from __future__ import annotations

from typing import List

from buildarr.config import RemoteMapEntry
from buildarr.types import BaseEnum
from typing_extensions import Self

from ..api import api_get, api_put
from ..secrets import SonarrSecrets
from .types import SonarrConfigBase


class FirstDayOfWeek(BaseEnum):
    """
    First day of the week enumeration for Sonarr.
    """

    sunday = 0
    monday = 1


class WeekColumnHeader(BaseEnum):
    """
    Week column header enumeration for Sonarr.
    """

    month_first = "ddd M/D"
    month_first_padded = "ddd MM/DD"
    day_first = "ddd D/M"
    day_first_padded = "ddd DD/MM"


class ShortDateFormat(BaseEnum):
    """
    Short date format enumeration for Sonarr.
    """

    word_month_first = "MMM D YYYY"
    word_month_second = "DD MMM YYYY"
    slash_month_first = "MM/D/YYYY"
    slash_month_first_day_padded = "MM/DD/YYYY"
    slash_day_first = "DD/MM/YYYY"
    iso8601 = "YYYY-MM-DD"


class LongDateFormat(BaseEnum):
    """
    Long date format enumeration for SOnarr.
    """

    month_first = "dddd, MMMM D YYYY"
    day_first = "dddd, D MMMM YYYY"


class TimeFormat(BaseEnum):
    """
    Time format enumeration for Sonarr.
    """

    twelve_hour = "h(:mm)a"
    twentyfour_hour = "HH:mm"


class SonarrUISettingsConfig(SonarrConfigBase):
    """
    Sonarr user interface configuration can also be set directly from Buildarr.

    ```yaml
    sonarr:
      settings:
        ui:
          first_day_of_week: "monday"
          week_column_header: "day-first"
          short_date_format: "word-month-second"
          long_date_format: "day-first"
          time_format: "twentyfour-hour"
          show_relative_dates: true
          enable_color_impaired_mode: false
    ```
    """

    # Calendar
    first_day_of_week: FirstDayOfWeek = FirstDayOfWeek.sunday
    """
    The first day of the week that Sonarr will show in the calendar.

    Values:

    * `sunday` - Sunday
    * `monday` - Monday
    """

    week_column_header: WeekColumnHeader = WeekColumnHeader.month_first
    """
    The format of the date in columns when "Week" is the active view in the calendar.

    Values:

    * `month-first` - Print month first (e.g. Tue 3/25)
    * `month-first-padded` - Print month first with padded numbers (e.g. Tue 03/25)
    * `day-first` - Print day first with padded numbers (e.g. Tue 25/3)
    * `day-first-padded` - Print day first with padded numbers (e.g. Tue 25/03)
    """

    # Dates
    short_date_format: ShortDateFormat = ShortDateFormat.word_month_first
    """
    The format of short dates in the user interface.

    Values:

    * `word-month-first` - Month as word, print month first (e.g. Mar 4 2014)
    * `word-month-second` - Month as word, print month second (e.g. 4 Mar 2014)
    * `slash-month-first` - Slash-separated date, print month first (e.g. 03/4/2014)
    * `slash-month-first-padded` - Slash-separated date, print month first (e.g. 03/04/2014)
    * `slash-day-first` - Slash-separated date, print day first (e.g. 04/03/2014)
    * `iso8601` - ISO-8601 date (e.g. 2014-03-04)
    """

    long_date_format: LongDateFormat = LongDateFormat.month_first
    """
    The format of long dates in the user interface.

    Values:

    * `month-first` - Print month first (e.g. Tuesday, March 4 2014)
    * `day-first` - Print day first (e.g. Tuesday, 4 March 2014)
    """

    time_format: TimeFormat = TimeFormat.twelve_hour
    """
    The format of time in the user information.

    Values:

    * `twelve-hour` - 12-hour time (e.g. 5pm/5:30pm)
    * `twentyfour-hour` - 24-hour time (e.g. 17:00/17:30)
    """

    show_relative_dates: bool = True
    """
    When set to `True`, Sonarr will show relative dates (e.g. today, yesterday)
    instead of absolute dates (e.g. Monday, Tuesday ...).
    """

    # Style
    enable_color_impaired_mode: bool = False
    """
    Enable an altered view style to allow colour-impaired users to better distinguish
    colour-coded information.
    """

    _remote_map: List[RemoteMapEntry] = [
        ("first_day_of_week", "firstDayOfWeek", {}),
        ("week_column_header", "calendarWeekColumnHeader", {}),
        ("short_date_format", "shortDateFormat", {}),
        ("long_date_format", "longDateFormat", {}),
        ("time_format", "timeFormat", {}),
        ("show_relative_dates", "showRelativeDates", {}),
        ("enable_color_impaired_mode", "enableColorImpairedMode", {}),
    ]

    @classmethod
    def from_remote(cls, secrets: SonarrSecrets) -> Self:
        return cls(
            **cls.get_local_attrs(
                remote_map=cls._remote_map,
                remote_attrs=api_get(secrets, "/api/v3/config/ui"),
            ),
        )

    def update_remote(
        self,
        tree: str,
        secrets: SonarrSecrets,
        remote: Self,
        check_unmanaged: bool = False,
    ) -> bool:
        updated, remote_attrs = self.get_update_remote_attrs(
            tree=tree,
            remote=remote,
            remote_map=self._remote_map,
            check_unmanaged=check_unmanaged,
        )
        if updated:
            config_id = api_get(secrets, "/api/v3/config/ui")["id"]
            api_put(
                secrets,
                f"/api/v3/config/ui/{config_id}",
                {"id": config_id, **remote_attrs},
            )
            return True
        return False



# Copyright (C) 2023 Callum Dickinson
#
# Buildarr is free software: you can redistribute it and/or modify it under the terms of the
# GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# Buildarr is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Buildarr.
# If not, see <https://www.gnu.org/licenses/>.


"""
Sonarr plugin CLI commands.
"""


from __future__ import annotations

import functools

from getpass import getpass
from urllib.parse import urlparse

import click
import click_params  # type: ignore[import]

from .config import SonarrInstanceConfig
from .manager import SonarrManager
from .secrets import SonarrSecrets

HOSTNAME_PORT_TUPLE_LENGTH = 2


@click.group(help="Sonarr instance ad-hoc commands.")
def sonarr():
    """
    Sonarr instance ad-hoc commands.
    """

    pass


@sonarr.command(
    help=(
        "Dump configuration from a remote Sonarr instance.\n\n"
        "The configuration is dumped to standard output in Buildarr-compatible YAML format."
    ),
)
@click.argument("url", type=click_params.URL)
@click.option(
    "-k",
    "--api-key",
    "api_key",
    metavar="API-KEY",
    default=functools.partial(getpass, "Sonarr instance API key: "),
    help="API key of the Sonarr instance. The user will be prompted if undefined.",
)
def dump_config(url: str, api_key: str) -> int:
    """
    Dump configuration from a remote Sonarr instance.
    The configuration is dumped to standard output in Buildarr-compatible YAML format.
    """

    url_obj = urlparse(url)
    protocol = url_obj.scheme
    hostname_port = url_obj.netloc.split(":", 1)
    hostname = hostname_port[0]
    port = (
        int(hostname_port[1])
        if len(hostname_port) == HOSTNAME_PORT_TUPLE_LENGTH
        else (443 if protocol == "https" else 80)
    )

    click.echo(
        SonarrManager()
        .from_remote(
            instance_config=SonarrInstanceConfig(hostname=hostname, port=port, protocol=protocol),
            secrets=SonarrSecrets(
                hostname=hostname,
                port=port,
                protocol=protocol,
                api_key=api_key,
            ),
        )
        .yaml(exclude_unset=True),
    )

    return 0



# Copyright (C) 2023 Callum Dickinson
#
# Buildarr is free software: you can redistribute it and/or modify it under the terms of the
# GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# Buildarr is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Buildarr.
# If not, see <https://www.gnu.org/licenses/>.


"""
Sonarr plugin delay profile configuration.
"""


from __future__ import annotations

from logging import getLogger
from typing import Any, Dict, List, Mapping, Set

from buildarr.config import RemoteMapEntry
from buildarr.types import BaseEnum, NonEmptyStr
from pydantic import Field
from typing_extensions import Self

from ...api import api_delete, api_get, api_post, api_put
from ...secrets import SonarrSecrets
from ..types import SonarrConfigBase

logger = getLogger(__name__)


class PreferredProtocol(BaseEnum):
    """
    Enumeration for enabled and preferred protocols in delay profiles.
    """

    usenet_prefer = 0
    torrent_prefer = 1
    usenet_only = 2
    torrent_only = 3

    @classmethod
    def get_from_params(
        cls,
        usenet_enabled: bool,
        torrent_enabled: bool,
        preferred_protocol: str,
    ) -> PreferredProtocol:
        return {
            (True, True, "usenet"): PreferredProtocol.usenet_prefer,
            (True, True, "unknown"): PreferredProtocol.usenet_prefer,
            (True, True, "torrent"): PreferredProtocol.torrent_prefer,
            (True, False, "usenet"): PreferredProtocol.usenet_only,
            (True, False, "unknown"): PreferredProtocol.usenet_only,
            (False, True, "torrent"): PreferredProtocol.torrent_only,
            (False, True, "unknown"): PreferredProtocol.torrent_only,
        }[(usenet_enabled, torrent_enabled, preferred_protocol)]

    @property
    def usenet_enabled(self) -> bool:
        return self in (
            PreferredProtocol.usenet_only,
            PreferredProtocol.usenet_prefer,
            PreferredProtocol.torrent_prefer,
        )

    @property
    def torrent_enabled(self) -> bool:
        return self in (
            PreferredProtocol.torrent_only,
            PreferredProtocol.torrent_prefer,
            PreferredProtocol.usenet_prefer,
        )

    @property
    def preferred_protocol(self) -> str:
        return (
            "torrent"
            if self in (PreferredProtocol.torrent_prefer, PreferredProtocol.torrent_only)
            else "usenet"
        )


class DelayProfile(SonarrConfigBase):
    """
    Delay profiles are defined as an ordered list of objects.

    A preferred protocol must be specified for all delay profiles.
    Tags must be defined on all except the final profile (the default profile),
    where tags must not be defined.

    ```yaml
    ...
      delay_profiles:
        definitions:
          # Ordered in priority, highest priority first.
          - preferred_protocol: "usenet-prefer" # Required
            usenet_delay: 0
            torrent_delay: 0
            bypass_if_highest_quality: true
            tags:
              - "tv-shows"
            # Add additional delay profiles here as needed.
            ...
            # Default delay profile goes last, and MUST be defined
            # if you have defined any other delay profiles.
            - preferred_protocol: "torrent-prefer" # Required
              usenet_delay: 1440
              torrent_delay: 1440
              bypass_if_highest_quality: false
              # Tags will be ignored for default delay profile.
    ```
    """

    preferred_protocol: PreferredProtocol
    """
    Choose which protocol(s) to use and which one is preferred
    when choosing between otherwise equal releases.

    Values:

    * `usenet-prefer` (Prefer Usenet)
    * `torrent-prefer` (Prefer Torrent)
    * `usenet-only` (Only Usenet)
    * `torrent-only` (Only Torrent)
    """

    usenet_delay: int = Field(0, ge=0)  # minutes
    """
    Delay (in minutes) to wait before grabbing a release from Usenet.
    """

    torrent_delay: int = Field(0, ge=0)  # minutes
    """
    Delay (in minutes) to wait before grabbing a torrent.
    """

    bypass_if_highest_quality: bool = False
    """
    Bypass the delay if a found release is the highest quality allowed
    in the quality profile that applies to it, and uses the preferred protocol
    as defined in this delay profile.
    """

    tags: Set[NonEmptyStr] = set()
    """
    Tags to assign to this delay profile.

    This delay profile will apply to series with at least one matching tag.
    """

    @classmethod
    def _get_remote_map(
        cls,
        tag_ids: Mapping[str, int],
    ) -> List[RemoteMapEntry]:
        return [
            (
                "preferred_protocol",
                "enableUsenet",
                {
                    "root_decoder": lambda vs: PreferredProtocol.get_from_params(
                        usenet_enabled=vs["enableUsenet"],
                        torrent_enabled=vs["enableTorrent"],
                        preferred_protocol=vs["preferredProtocol"],
                    ),
                    "encoder": lambda v: v.usenet_enabled,
                },
            ),
            (
                "preferred_protocol",
                "enableTorrent",
                {
                    "root_decoder": lambda vs: PreferredProtocol.get_from_params(
                        usenet_enabled=vs["enableUsenet"],
                        torrent_enabled=vs["enableTorrent"],
                        preferred_protocol=vs["preferredProtocol"],
                    ),
                    "encoder": lambda v: v.torrent_enabled,
                },
            ),
            (
                "preferred_protocol",
                "preferredProtocol",
                {
                    "root_decoder": lambda vs: PreferredProtocol.get_from_params(
                        usenet_enabled=vs["enableUsenet"],
                        torrent_enabled=vs["enableTorrent"],
                        preferred_protocol=vs["preferredProtocol"],
                    ),
                    "encoder": lambda v: v.preferred_protocol,
                },
            ),
            ("usenet_delay", "usenetDelay", {}),
            ("torrent_delay", "torrentDelay", {}),
            ("bypass_if_highest_quality", "bypassIfHighestQuality", {}),
            (
                "tags",
                "tags",
                {
                    "decoder": lambda v: set(
                        (tag for tag, tag_id in tag_ids.items() if tag_id in v),
                    ),
                    "encoder": lambda v: sorted(tag_ids[tag] for tag in v),
                },
            ),
        ]

    @classmethod
    def _from_remote(cls, tag_ids: Mapping[str, int], remote_attrs: Mapping[str, Any]) -> Self:
        return cls(
            **cls.get_local_attrs(cls._get_remote_map(tag_ids), remote_attrs),
        )

    def _create_remote(
        self,
        tree: str,
        secrets: SonarrSecrets,
        tag_ids: Mapping[str, int],
        order: int,
    ) -> None:
        api_post(
            secrets,
            "/api/v3/delayprofile",
            {
                "order": order,
                **self.get_create_remote_attrs(tree, self._get_remote_map(tag_ids)),
            },
        )

    def _update_remote(
        self,
        tree: str,
        secrets: SonarrSecrets,
        remote: DelayProfile,
        tag_ids: Mapping[str, int],
        profile_id: int,
        order: int,
    ) -> bool:
        changed, remote_attrs = self.get_update_remote_attrs(
            tree,
            remote,
            self._get_remote_map(tag_ids),
            check_unmanaged=True,
            set_unchanged=True,
        )
        if changed:
            api_put(
                secrets,
                f"/api/v3/delayprofile/{profile_id}",
                {"id": profile_id, "order": order, **remote_attrs},
            )
            return True
        return False

    def _delete_remote(
        self,
        tree: str,
        secrets: SonarrSecrets,
        profile_id: int,
    ) -> None:
        logger.info("%s: (...) -> (deleted)", tree)
        api_delete(secrets, f"/api/v3/delayprofile/{profile_id}")


class SonarrDelayProfilesSettingsConfig(SonarrConfigBase):
    """
    Configuration parameters for controlling how Buildarr handles delay profiles.
    """

    delete_unmanaged = False
    """
    Controls how Buildarr manages existing delay profiles in Sonarr when no delay profiles
    are defined in Buildarr.

    When set to `True` and there are no delay profiles defined in Buildarr,
    delete all delay profiles except the default delay profile (which can't be deleted).

    When set to `False` and there are no delay profiles defined in Buildarr,
    do not modify the existing delay profiles in Sonarr at all.

    Due to the unique way delay profiles are structured, when they are defined in Buildarr,
    they always overwrite the existing delay profiles on the remote Sonarr instance
    and configure it exactly as laid out in Buildarr, irrespective of this value.

    If unsure, leave this value set to `False`.
    """

    definitions: List[DelayProfile] = []
    """
    Define delay profiles to configure on Sonarr here.

    The final delay profile in the list is assumed to be the default delay profile.
    """

    # TODO: add a validator that checks that all definitions except the last one have tags,
    #       and the last one has no tags.

    @classmethod
    def from_remote(cls, secrets: SonarrSecrets) -> Self:
        profiles: List[Dict[str, Any]] = sorted(
            api_get(secrets, "/api/v3/delayprofile"),
            key=lambda p: p["order"],
            reverse=True,
        )
        tag_ids: Dict[str, int] = (
            {tag["label"]: tag["id"] for tag in api_get(secrets, "/api/v3/tag")}
            if any(profile["tags"] for profile in profiles)
            else {}
        )
        return cls(
            definitions=[DelayProfile._from_remote(tag_ids, profile) for profile in profiles],
        )

    def update_remote(
        self,
        tree: str,
        secrets: SonarrSecrets,
        remote: Self,
        check_unmanaged: bool = False,
    ) -> bool:
        if not self.delete_unmanaged and "definitions" not in self.__fields_set__:
            # TODO: Print current delay profile structure.
            if remote.definitions:
                logger.debug("%s.definitions: [...] (unmanaged)", tree)
            else:
                logger.debug("%s.definitions: [] (unmanaged)", tree)
            return False
        #
        changed = False
        #
        profile_ids: Dict[int, int] = {
            profile["order"]: profile["id"] for profile in api_get(secrets, "/api/v3/delayprofile")
        }
        tag_ids: Dict[str, int] = (
            {tag["label"]: tag["id"] for tag in api_get(secrets, "/api/v3/tag")}
            if any(profile.tags for profile in self.definitions)
            or any(profile.tags for profile in remote.definitions)
            else {}
        )
        #
        # Order: update already exist -> create/delete
        for order in range(max(len(self.definitions), len(remote.definitions))):
            local_index = len(self.definitions) - 1 - order
            remote_index = len(remote.definitions) - 1 - order
            # If the index is negative, then there are more delay profiles
            # on the remote than there are defined in the local configuration.
            # Delete those extra delay profiles from the remote.
            if local_index < 0:
                remote.definitions[remote_index]._delete_remote(
                    tree=f"{tree}.definitions[{local_index}]",
                    secrets=secrets,
                    profile_id=profile_ids[order],
                )
                changed = True
            # If the current index (order) is one that does not exist on the remote, create it.
            elif remote_index < 0:
                self.definitions[local_index]._create_remote(
                    tree=f"{tree}.definitions[{local_index}]",
                    secrets=secrets,
                    tag_ids=tag_ids,
                    order=order,
                )
                changed = True
            # If none of the above conditions checked out, then the current index exists
            # in both the local configuration and the remote.
            # Check and update those delay profiles.
            elif self.definitions[local_index]._update_remote(
                tree=f"{tree}.definitions[{local_index}]",
                secrets=secrets,
                remote=remote.definitions[remote_index],
                tag_ids=tag_ids,
                profile_id=profile_ids[order],
                order=order,
            ):
                changed = True
        #
        return changed



# Copyright (C) 2023 Callum Dickinson
#
# Buildarr is free software: you can redistribute it and/or modify it under the terms of the
# GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# Buildarr is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Buildarr.
# If not, see <https://www.gnu.org/licenses/>.


"""
Sonarr plugin configuration utility classes and functions.
"""


from __future__ import annotations

from typing import TYPE_CHECKING

from buildarr.config import ConfigBase
from pydantic import ConstrainedStr

if TYPE_CHECKING:
    from ..secrets import SonarrSecrets

    class SonarrConfigBase(ConfigBase[SonarrSecrets]):
        ...

else:

    class SonarrConfigBase(ConfigBase):
        ...


class TraktAuthUser(ConstrainedStr):
    """
    Constrained string type to make the Trakt auth user case-insensitive.
    """

    min_length = 1
    to_lower = True



# Copyright (C) 2023 Callum Dickinson
#
# Buildarr is free software: you can redistribute it and/or modify it under the terms of the
# GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# Buildarr is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Buildarr.
# If not, see <https://www.gnu.org/licenses/>.


"""
Sonarr plugin configuration.
"""


from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Optional

from buildarr.config import ConfigPlugin
from buildarr.types import NonEmptyStr, Port
from typing_extensions import Self

from ..api import api_get
from ..types import SonarrApiKey, SonarrProtocol
from .connect import SonarrConnectSettingsConfig
from .download_clients import SonarrDownloadClientsSettingsConfig
from .general import SonarrGeneralSettingsConfig
from .import_lists import SonarrImportListsSettingsConfig
from .indexers import SonarrIndexersSettingsConfig
from .media_management import SonarrMediaManagementSettingsConfig
from .metadata import SonarrMetadataSettingsConfig
from .profiles import SonarrProfilesSettingsConfig
from .quality import SonarrQualitySettingsConfig
from .tags import SonarrTagsSettingsConfig
from .types import SonarrConfigBase
from .ui import SonarrUISettingsConfig

if TYPE_CHECKING:
    from ..secrets import SonarrSecrets

    class _SonarrInstanceConfig(ConfigPlugin[SonarrSecrets]):
        ...

else:

    class _SonarrInstanceConfig(ConfigPlugin):
        ...


class SonarrSettingsConfig(SonarrConfigBase):
    """
    Sonarr settings, used to configure a remote Sonarr instance.
    """

    #: Media management settings.
    media_management = SonarrMediaManagementSettingsConfig()

    #: Profile settings (quality, language, delay, release profiles).
    profiles = SonarrProfilesSettingsConfig()

    #: Quality definition settings.
    quality = SonarrQualitySettingsConfig()

    #: Usenet/BitTorrent indexer settings.
    indexers = SonarrIndexersSettingsConfig()

    #: Usenet/BitTorrent download client settings.
    download_clients = SonarrDownloadClientsSettingsConfig()

    #: Import list settings.
    import_lists = SonarrImportListsSettingsConfig()

    #: Notification connection settings.
    connect = SonarrConnectSettingsConfig()

    #: Metadata file settings.
    metadata = SonarrMetadataSettingsConfig()

    #: Tag settings.
    tags = SonarrTagsSettingsConfig()

    #: General instance settings.
    general = SonarrGeneralSettingsConfig()

    #: User interface settings.
    ui = SonarrUISettingsConfig()

    def update_remote(
        self,
        tree: str,
        secrets: SonarrSecrets,
        remote: Self,
        check_unmanaged: bool = False,
    ) -> bool:
        # Overload base function to guarantee execution order of section updates.
        # 1. Tags must be created before everything else, and destroyed after they
        #    are no longer referenced elsewhere.
        # 2. Qualities must be updated before quality profiles.
        # 3. Indexers must be created before release profiles, and destroyed after they
        #    are no longer referenced by them.
        return any(
            [
                self.tags.update_remote(
                    f"{tree}.tags",
                    secrets,
                    remote.tags,
                    check_unmanaged=check_unmanaged,
                ),
                self.quality.update_remote(
                    f"{tree}.quality",
                    secrets,
                    remote.quality,
                    check_unmanaged=check_unmanaged,
                ),
                self.indexers.update_remote(
                    f"{tree}.indexers",
                    secrets,
                    remote.indexers,
                    check_unmanaged=check_unmanaged,
                ),
                self.media_management.update_remote(
                    f"{tree}.media_management",
                    secrets,
                    remote.media_management,
                    check_unmanaged=check_unmanaged,
                ),
                self.profiles.update_remote(
                    f"{tree}.profiles",
                    secrets,
                    remote.profiles,
                    check_unmanaged=check_unmanaged,
                ),
                self.download_clients.update_remote(
                    f"{tree}.download_clients",
                    secrets,
                    remote.download_clients,
                    check_unmanaged=check_unmanaged,
                ),
                self.import_lists.update_remote(
                    f"{tree}.import_lists",
                    secrets,
                    remote.import_lists,
                    check_unmanaged=check_unmanaged,
                ),
                self.connect.update_remote(
                    f"{tree}.connect",
                    secrets,
                    remote.connect,
                    check_unmanaged=check_unmanaged,
                ),
                self.metadata.update_remote(
                    f"{tree}.metadata",
                    secrets,
                    remote.metadata,
                    check_unmanaged=check_unmanaged,
                ),
                self.general.update_remote(
                    f"{tree}.general",
                    secrets,
                    remote.general,
                    check_unmanaged=check_unmanaged,
                ),
                self.ui.update_remote(
                    f"{tree}.ui",
                    secrets,
                    remote.ui,
                    check_unmanaged=check_unmanaged,
                ),
                # TODO: destroy indexers
                # TODO: destroy tags
            ],
        )


class SonarrInstanceConfig(_SonarrInstanceConfig):
    """
    By default, Buildarr will look for a single instance at `http://sonarr:8989`.
    Most configurations are different, and to accommodate those, you can configure
    how Buildarr connects to individual Sonarr instances.

    Configuration of a single Sonarr instance:

    ```yaml
    sonarr:
      hostname: "sonarr.example.com"
      port: 8989
      protocol: "http"
      settings:
        ...
    ```

    Configuration of multiple instances:

    ```yaml
    sonarr:
      # Configuration and settings common to all instances.
      port: 8989
      settings:
        ...
      instances:
        # Sonarr instance 1-specific configuration.
        sonarr1:
          hostname: "sonarr1.example.com"
          settings:
            ...
        # Sonarr instance 2-specific configuration.
        sonarr2:
          hostname: "sonarr2.example.com"
          api_key: "..." # Explicitly define API key
          settings:
            ...
    ```
    """

    hostname: NonEmptyStr = "sonarr"  # type: ignore[assignment]
    """
    Hostname of the Sonarr instance to connect to.

    When defining a single instance using the global `sonarr` configuration block,
    the default hostname is `sonarr`.

    When using multiple instance-specific configurations, the default hostname
    is the name given to the instance in the `instances` attribute.

    ```yaml
    sonarr:
      instances:
        sonarr1: # <--- This becomes the default hostname
          ...
    ```
    """

    port: Port = 8989  # type: ignore[assignment]
    """
    Port number of the Sonarr instance to connect to.
    """

    protocol: SonarrProtocol = "http"  # type: ignore[assignment]
    """
    Communication protocol to use to connect to Sonarr.
    """

    api_key: Optional[SonarrApiKey] = None
    """
    API key to use to authenticate with the Sonarr instance.

    If undefined or set to `None`, automatically retrieve the API key.
    This can only be done on Sonarr instances with authentication disabled.
    """

    image: NonEmptyStr = "lscr.io/linuxserver/sonarr"  # type: ignore[assignment]
    """
    The default Docker image URI when generating a Docker Compose file.
    """

    version: Optional[str] = None
    """
    The expected version of the Sonarr instance.
    If undefined or set to `None`, the version is auto-detected.

    This value is also used when generating a Docker Compose file.
    When undefined or set to `None`, the version tag will be set to `latest`.
    """

    settings: SonarrSettingsConfig = SonarrSettingsConfig()
    """
    Sonarr settings.
    Configuration options for Sonarr itself are set within this structure.
    """

    @property
    def uses_trash_metadata(self) -> bool:
        """
        A flag determining whether or not this configuration uses TRaSH-Guides metadata.

        Returns:
            `True` if TRaSH-Guides metadata is used, otherwise `False`
        """
        if self.settings.quality.uses_trash_metadata:
            return True
        for release_profile in self.settings.profiles.release_profiles.definitions.values():
            if release_profile.uses_trash_metadata:
                return True
        return False

    def render_trash_metadata(self, trash_metadata_dir: Path) -> Self:
        """
        Read TRaSH-Guides metadata, and return a configuration object with all templates rendered.

        Args:
            trash_metadata_dir (Path): TRaSH-Guides metadata directory.

        Returns:
            Rendered configuration object
        """
        copy = self.copy(deep=True)
        copy._render_trash_metadata(trash_metadata_dir)
        return copy

    def _render_trash_metadata(self, trash_metadata_dir: Path) -> None:
        """
        Render configuration attributes obtained from TRaSH-Guides, in-place.

        Args:
            trash_metadata_dir (Path): TRaSH-Guides metadata directory.
        """
        for rp in self.settings.profiles.release_profiles.definitions.values():
            if rp.uses_trash_metadata:
                rp._render_trash_metadata(trash_metadata_dir)
        if self.settings.quality.uses_trash_metadata:
            self.settings.quality._render_trash_metadata(trash_metadata_dir)

    @classmethod
    def from_remote(cls, secrets: SonarrSecrets) -> Self:
        """
        Read configuration from a remote instance and return it as a configuration object.

        Args:
            secrets (SonarrSecrets): Instance host and secrets information

        Returns:
            Configuration object for remote instance
        """
        return cls(
            hostname=secrets.hostname,
            port=secrets.port,
            protocol=secrets.protocol,
            api_key=secrets.api_key,
            version=api_get(secrets, "/api/v3/system/status")["version"],
            settings=SonarrSettingsConfig.from_remote(secrets),
        )

    def to_compose_service(self, compose_version: str, service_name: str) -> Dict[str, Any]:
        return {
            "image": f"{self.image}:{self.version or 'latest'}",
            "volumes": {service_name: "/config"},
        }


class SonarrConfig(SonarrInstanceConfig):
    """
    Sonarr plugin global configuration class.
    """

    instances: Dict[str, SonarrInstanceConfig] = {}
    """
    Instance-specific Sonarr configuration.

    Can only be defined on the global `sonarr` configuration block.

    Globally specified configuration values apply to all instances.
    Configuration values specified on an instance-level take precedence at runtime.
    """

    @property
    def uses_trash_metadata(self) -> bool:
        """
        A flag determining whether or not this configuration uses TRaSH-Guides metadata.

        Returns:
            `True` if TRaSH-Guides metadata is used, otherwise `False`
        """
        for instance in self.instances.values():
            if instance.uses_trash_metadata:
                return True
        return super().uses_trash_metadata

    def render_trash_metadata(self, trash_metadata_dir: Path) -> Self:
        """
        Read TRaSH-Guides metadata, and return a configuration object with all templates rendered.

        Args:
            trash_metadata_dir (Path): TRaSH-Guides metadata directory.

        Returns:
            Rendered configuration object
        """
        copy = self.copy(deep=True)
        for instance in copy.instances.values():
            if instance.uses_trash_metadata:
                instance._render_trash_metadata(trash_metadata_dir)
        if self.uses_trash_metadata:
            copy._render_trash_metadata(trash_metadata_dir)
        return copy



# Copyright (C) 2023 Callum Dickinson
#
# Buildarr is free software: you can redistribute it and/or modify it under the terms of the
# GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# Buildarr is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Buildarr.
# If not, see <https://www.gnu.org/licenses/>.


"""
Sonarr plugin download client remote path mappings.
"""


from __future__ import annotations

from logging import getLogger
from typing import Any, Dict, List, Mapping, Tuple

from buildarr.config import RemoteMapEntry
from buildarr.types import BaseEnum, NonEmptyStr
from typing_extensions import Self

from ...api import api_delete, api_get, api_post, api_put
from ...secrets import SonarrSecrets
from ..types import SonarrConfigBase

logger = getLogger(__name__)


class Ensure(BaseEnum):
    """
    Resource 'ensure' value enumeration.

    Values:

    * `present` (ensure the resource is created if it does not exist)
    * `absent` (ensure the resource is deleted if it exists)
    """

    present = "present"
    absent = "absent"


class RemotePathMapping(SonarrConfigBase):
    """
    Remote path mapping definitions themselves are relatively simple.

    They can be configured with a desired `ensure` state, however, which
    Buildarr uses to control whether to create or delete the remote path mapping.
    Ensure this value is set appropriately.
    """

    host: NonEmptyStr
    """
    The name of the host, as specified for the remote download client.
    """

    remote_path: NonEmptyStr
    """
    Root path to the directory that the download client accesses.
    """

    local_path: NonEmptyStr
    """
    The path that Sonarr should use to access the remote path locally.
    """

    ensure: Ensure = Ensure.present
    """
    Desired state for this resource.

    If set to `present`, the resource is created on the remote instance if it does not exist.

    If set to `absent`, the resource will be destroyed on the remote instance if it exists.
    This takes effect even if the  `delete_unmanaged` is set to `False`
    for all remote path mappings.
    """

    _remote_map: List[RemoteMapEntry] = [
        ("host", "host", {}),
        ("remote_path", "remotePath", {}),
        ("local_path", "localPath", {}),
    ]

    @classmethod
    def _from_remote(cls, remote_attrs: Mapping[str, Any]) -> Self:
        return cls(**cls.get_local_attrs(cls._remote_map, remote_attrs))

    def _create_remote(self, tree: str, secrets: SonarrSecrets) -> None:
        api_post(
            secrets,
            "/api/v3/remotepathmapping",
            self.get_create_remote_attrs(tree, self._remote_map),
        )

    def _update_remote(
        self,
        tree: str,
        secrets: SonarrSecrets,
        remote: Self,
        remotepathmapping_id: int,
    ) -> bool:
        updated, remote_attrs = self.get_update_remote_attrs(
            tree,
            remote,
            self._remote_map,
            check_unmanaged=True,
            set_unchanged=True,
        )
        if updated:
            api_put(
                secrets,
                f"/api/v3/remotepathmapping/{remotepathmapping_id}",
                {"id": remotepathmapping_id, **remote_attrs},
            )
            return True
        return False

    def _delete_remote(
        self,
        tree: str,
        secrets: SonarrSecrets,
        remotepathmapping_id: int,
    ) -> None:
        logger.info("%s: (...) -> (deleted)", tree)
        api_delete(secrets, f"/api/v3/remotepathmapping/{remotepathmapping_id}")


class SonarrRemotePathMappingsSettingsConfig(SonarrConfigBase):
    """
    Remote path mappings are used to associate a path on a download client remote host
    with its associated path on the local Sonarr instance.

    The main use case for this is when Sonarr and the download client are not running
    on the same system, or when Docker is used to isolate these services and the
    mountpoints for media locations are not consistent between the containers.

    ```yaml
    sonarr:
      settings:
        download_clients:
          definitions:
            Transmission:
              type: "transmission"
              host: "transmission"
              port: 9091
          remote_path_mappings:
            delete_unmanaged: false # Optional
            definitions:
              - host: "transmission"
                remote_path: "/remote/path"
                local_path: "/local/path"
                ensure: "present" # Optional
    ```

    Remote path mappings can be difficult to properly configure.
    TRaSH-Guides provides an
    [excellent guide](https://trash-guides.info/Sonarr/Sonarr-remote-path-mapping/)
    that explains what they are for, and how to use them.
    """

    delete_unmanaged: bool = False
    """
    Automatically delete remote path mappings not configured in Buildarr.

    Deleting existing remote path mappings can cause problems with a running
    Sonarr instance. Think carefully before you enable this option.
    """

    # TODO: validator to ensure every mapping is unique
    definitions: List[RemotePathMapping] = []
    """
    Remote path mapping definitions.
    """

    @classmethod
    def _from_remote(cls, secrets: SonarrSecrets) -> Self:
        return cls(
            definitions=sorted(
                (
                    RemotePathMapping._from_remote(rpm)
                    for rpm in api_get(secrets, "/api/v3/remotepathmapping")
                ),
                key=lambda v: (v.host, v.remote_path, v.local_path),
            ),
        )

    def _update_remote(
        self,
        tree: str,
        secrets: SonarrSecrets,
        remote: Self,
    ) -> bool:
        # Track whether remote resource path mappings have been updated.
        changed = False
        # Get required resource IDs from the remote, and create
        # data structures.
        remote_rpm_ids: Dict[Tuple[str, str, str], int] = {
            (rpm["host"], rpm["remotePath"], rpm["localPath"]): rpm["id"]
            for rpm in api_get(secrets, "/api/v3/remotepathmapping")
        }
        local_rpms: Dict[Tuple[str, str, str], RemotePathMapping] = {
            (rpm.host, rpm.remote_path, rpm.local_path): rpm for rpm in self.definitions
        }
        remote_rpms: Dict[Tuple[str, str, str], RemotePathMapping] = {
            (rpm.host, rpm.remote_path, rpm.local_path): rpm for rpm in remote.definitions
        }
        # Handle managed remote path mappings.
        for i, rpm in enumerate(self.definitions):
            rpm_tree = f"{tree}.definitions[{i}]"
            rpm_tuple = (rpm.host, rpm.remote_path, rpm.local_path)
            # If the remote path mapping should exist, check that it does,
            # and if not, create it.
            if rpm.ensure == Ensure.present:
                if rpm_tuple in remote_rpms:
                    logger.debug("%s: %s (exists)", rpm_tree, repr(rpm))
                else:
                    logger.info("%s: %s -> (created)", rpm_tree, repr(rpm))
                    rpm._create_remote(tree=rpm_tree, secrets=secrets)
                    changed = True
            # If the remote path mapping should not exist, check that it does not
            # exist in the remote, and if it does, delete it.
            elif rpm_tuple in remote_rpms:
                logger.info("%s: %s -> (deleted)", rpm_tree, repr(rpm))
                rpm._delete_remote(
                    tree=rpm_tree,
                    secrets=secrets,
                    remotepathmapping_id=remote_rpm_ids[rpm_tuple],
                )
                changed = True
            else:
                logger.debug("%s: %s (does not exist)", rpm_tree, repr(rpm))
        # Handle unmanaged remote path mappings.
        # If `delete_unmanaged` is `True`, automatically delete them.
        j = -1
        for rpm in remote.definitions:
            rpm_tuple = (rpm.host, rpm.remote_path, rpm.local_path)
            if rpm_tuple not in local_rpms:
                rpm_tree = f"{tree}.definitions[{j}]"
                if self.delete_unmanaged:
                    logger.info("%s: %s -> (deleted)", rpm_tree, repr(rpm))
                    rpm._delete_remote(
                        tree=rpm_tree,
                        secrets=secrets,
                        remotepathmapping_id=remote_rpm_ids[rpm_tuple],
                    )
                    changed = True
                else:
                    logger.debug("%s: %s (unmanaged)", rpm_tree, repr(rpm))
                j -= 1
        # Return changed status.
        return changed



# Copyright (C) 2023 Callum Dickinson
#
# Buildarr is free software: you can redistribute it and/or modify it under the terms of the
# GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# Buildarr is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Buildarr.
# If not, see <https://www.gnu.org/licenses/>.


"""
Sonarr plugin quality profile configuration.
"""


from __future__ import annotations

from logging import getLogger
from typing import Any, Dict, List, Mapping, Optional, Sequence, Set, Union

from buildarr.config import RemoteMapEntry
from buildarr.types import NonEmptyStr
from pydantic import Field, validator
from typing_extensions import Annotated, Self

from ...api import api_delete, api_get, api_post, api_put
from ...secrets import SonarrSecrets
from ..types import SonarrConfigBase

logger = getLogger(__name__)


class QualityGroup(SonarrConfigBase):
    """
    Quality group.

    Allows groups of quality definitions to be given the same prorioty in qualtity profiles.
    """

    name: NonEmptyStr
    members: Set[NonEmptyStr] = Field(..., min_items=1)

    def encode(self, group_id: int, quality_definitions: Mapping[str, Any]) -> Dict[str, Any]:
        return {
            "id": group_id,
            "name": self.name,
            "allowed": True,
            "items": [
                _encode_quality_str(quality_definitions, member, True) for member in self.members
            ],
        }


class QualityProfile(SonarrConfigBase):
    """
    The main things to consider when creating a quality profile are
    what quality settings to enable, and how to prioritise each.

    ```yaml
    ...
      quality_profiles:
        SDTV:
          upgrades_allowed: true
          upgrade_until: "Bluray-1080p"
          qualities:
          - "Bluray-480p"
          - "DVD"
          - name: "WEB 480p"
            members:
              - "WEBDL-480p"
              - "WEBRip-480p"
          - "SDTV"
    ```

    In Buildarr, the quality listed first (at the top) is given the highest priority, with
    subsequent qualities given lower priority. Qualities not explicitly defined are
    disabled (not downloaded).

    Sonarr supports grouping multiple qualities together to give them the same priority.
    In Buildarr, these are expressed by giving a `name` to the group, and listing the qualities
    under the `members` attribute.

    For more insight into reasonable values for quality profiles,
    refer to these guides from [WikiArr](https://wiki.servarr.com/sonarr/settings#quality-profiles)
    and TRaSH-Guides ([WEB-DL](https://trash-guides.info/Sonarr/Sonarr-Release-Profile-RegEx/),
    [anime](https://trash-guides.info/Sonarr/Sonarr-Release-Profile-RegEx-Anime/)).
    """

    upgrades_allowed: bool = False
    """
    Enable automatic upgrading if a higher quality version of the media file becomes available.

    If disabled, media files will not be upgraded after they have been downloaded.
    """

    qualities: Annotated[List[Union[NonEmptyStr, QualityGroup]], Field(min_items=1)]
    """
    The qualities to enable downloading episodes for. The order determines the priority
    (highest priority first, lowest priority last).

    Individual qualities can be specified using the name (e.g. `Bluray-480p`).

    Qualities can also be grouped together in a structure to give them the same
    priority level. A new version of the episode will not be downloaded if it is
    at least one of the qualities listed in the group, until a higher quality
    version is found.

    ```yaml
    ...
      qualities:
        - name: "WEB 480p"
          members:
            - "WEBDL-480p"
            - "WEBRip-480p"
    ```

    At least one quality must be specified.
    """

    upgrade_until: Optional[NonEmptyStr] = None
    """
    The maximum quality level to upgrade an episode to.
    For a quality group, specify the group name.

    Once this quality is reached Sonarr will no longer download episodes.

    This attribute is required if `upgrades_allowed` is set to `True`.
    """

    @validator("qualities")
    def validate_qualities(
        cls,
        value: List[Union[str, QualityGroup]],
    ) -> List[Union[str, QualityGroup]]:
        quality_name_map: Dict[str, Union[str, QualityGroup]] = {}
        for quality in value:
            for name in quality.members if isinstance(quality, QualityGroup) else [quality]:
                if name in quality_name_map:
                    error_message = f"duplicate entries of quality value '{name}' exist ("
                    other = quality_name_map[name]
                    if isinstance(quality, str) and isinstance(quality, str):
                        error_message += "both are non-grouped quality values"
                    else:
                        error_message += (
                            f"one as part of quality group '{quality.name}', "
                            if isinstance(quality, QualityGroup)
                            else "one as a non-grouped quality value, "
                        )
                        error_message += (
                            f"another as part of quality group '{other.name}'"
                            if isinstance(other, QualityGroup)
                            else "another as a non-grouped quality value"
                        )
                    error_message += ")"
                    raise ValueError(error_message)
                quality_name_map[name] = quality
        return value

    @validator("upgrade_until")
    def validate_upgrade_until(
        cls,
        value: Optional[str],
        values: Dict[str, Any],
    ) -> Optional[str]:
        try:
            upgrades_allowed: bool = values["upgrades_allowed"]
            qualities: Sequence[Union[str, QualityGroup]] = values["qualities"]
        except KeyError:
            return value
        # If `upgrades_allowed` is `False`, set `upgrade_until` to `None`
        # to make sure Buildarr ignores whatever it is currently set to
        # on the remote instance.
        if not upgrades_allowed:
            return None
        # Subsequent checks now assume that `upgrades_allowed` is `True`,
        # this parameter is required and defined to a valid value.
        if not value:
            raise ValueError("required if 'upgrades_allowed' is True")
        for quality in qualities:
            quality_name = quality.name if isinstance(quality, QualityGroup) else quality
            if value == quality_name:
                break
        else:
            raise ValueError("must be set to a value enabled in 'qualities'")
        return value

    @classmethod
    def _get_remote_map(
        cls,
        quality_definitions: Mapping[str, Mapping[str, Any]] = {},
        group_ids: Mapping[str, int] = {},
    ) -> List[RemoteMapEntry]:
        return [
            ("upgrades_allowed", "upgradeAllowed", {}),
            (
                "upgrade_until",
                "cutoff",
                {
                    "root_decoder": lambda vs: cls._upgrade_until_decoder(
                        items=vs["items"],
                        cutoff=vs["cutoff"],
                    ),
                    "root_encoder": lambda vs: cls._upgrade_until_encoder(
                        quality_definitions=quality_definitions,
                        group_ids=group_ids,
                        qualities=vs.qualities,
                        upgrade_until=vs.upgrade_until,
                    ),
                },
            ),
            (
                "qualities",
                "items",
                {
                    "decoder": lambda v: _decode_qualities(v),
                    "encoder": lambda v: _encode_qualities(quality_definitions, group_ids, v),
                },
            ),
        ]

    @classmethod
    def _upgrade_until_decoder(
        cls,
        items: Sequence[Mapping[str, Any]],
        cutoff: int,
    ) -> str:
        for quality_item in items:
            quality: Mapping[str, Any] = (
                quality_item  # Quality group
                if "id" in quality_item
                else quality_item["quality"]  # Quality definition (singular)
            )
            if quality["id"] == cutoff:
                return quality["name"]
        raise RuntimeError(
            "Inconsistent Sonarr instance state: "
            f"'cutoff' quality ID {cutoff} not found in 'items': {items}",
        )

    @classmethod
    def _upgrade_until_encoder(
        cls,
        quality_definitions: Mapping[str, Mapping[str, Any]],
        group_ids: Mapping[str, int],
        qualities: Sequence[Union[str, QualityGroup]],
        upgrade_until: Optional[str],
    ) -> int:
        if not upgrade_until:
            quality = qualities[0]
            return (
                group_ids[quality.name]
                if isinstance(quality, QualityGroup)
                else quality_definitions[quality]["id"]
            )
        return (
            group_ids[upgrade_until]
            if upgrade_until in group_ids
            else quality_definitions[upgrade_until]["id"]
        )

    @classmethod
    def _from_remote(cls, remote_attrs: Mapping[str, Any]) -> Self:
        return cls(**cls.get_local_attrs(cls._get_remote_map(), remote_attrs))

    def _create_remote(
        self,
        tree: str,
        secrets: SonarrSecrets,
        profile_name: str,
        quality_definitions: Mapping[str, Mapping[str, Any]],
    ) -> None:
        group_ids: Dict[str, int] = {
            quality_group.name: (1000 + i)
            for i, quality_group in enumerate(
                [q for q in self.qualities if isinstance(q, QualityGroup)],
                1,
            )
        }
        api_post(
            secrets,
            "/api/v3/qualityprofile",
            {
                "name": profile_name,
                **self.get_create_remote_attrs(
                    tree,
                    self._get_remote_map(quality_definitions, group_ids),
                ),
            },
        )

    def _update_remote(
        self,
        tree: str,
        secrets: SonarrSecrets,
        remote: Self,
        profile_id: int,
        profile_name: str,
        quality_definitions: Mapping[str, Mapping[str, Any]],
    ) -> bool:
        group_ids: Dict[str, int] = {
            quality_group.name: (1000 + i)
            for i, quality_group in enumerate(
                [q for q in self.qualities if isinstance(q, QualityGroup)],
                1,
            )
        }
        changed, remote_attrs = self.get_update_remote_attrs(
            tree,
            remote,
            self._get_remote_map(quality_definitions, group_ids),
            check_unmanaged=True,
            set_unchanged=True,
        )
        if changed:
            api_put(
                secrets,
                f"/api/v3/qualityprofile/{profile_id}",
                {"id": profile_id, "name": profile_name, **remote_attrs},
            )
            return True
        return False

    def _delete_remote(self, tree: str, secrets: SonarrSecrets, profile_id: int) -> None:
        logger.info("%s: (...) -> (deleted)", tree)
        api_delete(secrets, f"/api/v3/qualityprofile/{profile_id}")


class SonarrQualityProfilesSettingsConfig(SonarrConfigBase):
    """
    Configuration parameters for controlling how Buildarr handles quality profiles.
    """

    delete_unmanaged: bool = False
    """
    Automatically delete quality profiles not defined in Buildarr.

    Out of the box Sonarr provides some pre-defined quality profiles.
    Take care when enabling this option, as those will also be deleted.
    """

    definitions: Dict[str, QualityProfile] = {}
    """
    Define quality profiles to configure on Sonarr here.

    If there are no quality profiles defined and `delete_unmanaged` is `False`,
    Buildarr will not modify existing quality profiles, but if `delete_unmanaged` is `True`,
    **Buildarr will delete all existing profiles. Be careful when using `delete_unmanaged`.**
    """

    @classmethod
    def from_remote(cls, secrets: SonarrSecrets) -> Self:
        return cls(
            definitions={
                profile["name"]: QualityProfile._from_remote(profile)
                for profile in api_get(secrets, "/api/v3/qualityprofile")
            },
        )

    def update_remote(
        self,
        tree: str,
        secrets: SonarrSecrets,
        remote: Self,
        check_unmanaged: bool = False,
    ) -> bool:
        #
        changed = False
        #
        profile_ids: Dict[str, int] = {
            profile_json["name"]: profile_json["id"]
            for profile_json in api_get(secrets, "/api/v3/qualityprofile")
        }
        quality_definitions: Dict[str, Dict[str, Any]] = {
            quality_json["title"]: quality_json["quality"]
            for quality_json in sorted(
                api_get(secrets, "/api/v3/qualitydefinition"),
                key=lambda q: q["weight"],
                reverse=True,
            )
        }
        #
        for profile_name, profile in self.definitions.items():
            profile_tree = f"{tree}.definitions[{repr(profile_name)}]"
            #
            if profile_name not in remote.definitions:
                profile._create_remote(
                    tree=profile_tree,
                    secrets=secrets,
                    profile_name=profile_name,
                    quality_definitions=quality_definitions,
                )
                changed = True
            #
            elif profile._update_remote(
                tree=profile_tree,
                secrets=secrets,
                remote=remote.definitions[profile_name],
                profile_id=profile_ids[profile_name],
                profile_name=profile_name,
                quality_definitions=quality_definitions,
            ):
                changed = True
        #
        for profile_name, profile in remote.definitions.items():
            if profile_name not in self.definitions:
                profile_tree = f"{tree}.definitions[{repr(profile_name)}]"
                if self.delete_unmanaged:
                    profile._delete_remote(
                        tree=profile_tree,
                        secrets=secrets,
                        profile_id=profile_ids[profile_name],
                    )
                    changed = True
                else:
                    logger.debug("%s: (...) (unmanaged)", profile_tree)
        #
        return changed


def _decode_qualities(value: Sequence[Mapping[str, Any]]) -> List[Union[str, QualityGroup]]:
    return [
        (
            QualityGroup(
                name=quality["name"],
                members=[member["quality"]["name"] for member in quality["items"]],
            )
            if quality["items"]
            else quality["quality"]["name"]
        )
        for quality in reversed(value)
        if quality["allowed"]
    ]


def _encode_qualities(
    quality_definitions: Mapping[str, Mapping[str, Any]],
    group_ids: Mapping[str, int],
    qualities: List[Union[str, QualityGroup]],
) -> List[Dict[str, Any]]:
    qualities_json: List[Dict[str, Any]] = []
    enabled_qualities: Set[str] = set()

    for quality in qualities:
        if isinstance(quality, QualityGroup):
            qualities_json.append(quality.encode(group_ids[quality.name], quality_definitions))
            for member in quality.members:
                enabled_qualities.add(member)
        else:
            qualities_json.append(_encode_quality_str(quality_definitions, quality, True))
            enabled_qualities.add(quality)

    for quality_name in quality_definitions.keys():
        if quality_name not in enabled_qualities:
            qualities_json.append(_encode_quality_str(quality_definitions, quality_name, False))

    return list(reversed(qualities_json))


def _encode_quality_str(
    quality_definitions: Mapping[str, Mapping[str, Any]],
    quality_name: str,
    allowed: bool,
) -> Dict[str, Any]:
    return {"quality": quality_definitions[quality_name], "items": [], "allowed": allowed}



# Copyright (C) 2023 Callum Dickinson
#
# Buildarr is free software: you can redistribute it and/or modify it under the terms of the
# GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# Buildarr is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Buildarr.
# If not, see <https://www.gnu.org/licenses/>.


"""
Sonarr plugin configuration utility classes and functions.
"""


from __future__ import annotations

from datetime import datetime, timezone


def trakt_expires_encoder(dt: datetime) -> str:
    """
    Trakt OAuth token `expires` field API value encoder.

    Converts a naive or timezone-aware `datetime` object to a ISO-8601 string
    in UTC time with a Zulu timezone marker.

    If the original `datetime` object is naive, assume it is already in UTC
    and do not convert.

    Args:
        dt (datetime): Object to encode

    Returns:
        ISO-8601 date-time string in UTC with Zulu timezone marker
    """

    if dt.tzinfo == timezone.utc:
        dt_aware = dt
    elif dt.tzinfo is not None:
        dt_aware = dt.astimezone(timezone.utc)
    else:
        dt_aware = dt.replace(tzinfo=timezone.utc)

    return dt_aware.isoformat().replace("+00:00", "Z")



# Copyright (C) 2023 Callum Dickinson
#
# Buildarr is free software: you can redistribute it and/or modify it under the terms of the
# GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
#
# Buildarr is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
# See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with Buildarr.
# If not, see <https://www.gnu.org/licenses/>.


"""
Sonarr plugin API functions.
"""


from __future__ import annotations

import json
import re

from datetime import datetime, timezone
from http import HTTPStatus
from logging import getLogger
from typing import TYPE_CHECKING

import json5  # type: ignore[import]
import requests

from buildarr.state import state

from .exceptions import SonarrAPIError

if TYPE_CHECKING:
    from typing import Any, Dict, Mapping, Optional

    from .secrets import SonarrSecrets


logger = getLogger(__name__)

INITIALIZE_JS_RES_PATTERN = re.compile(r"(?s)^window\.Sonarr = ({.*});$")


def get_initialize_js(host_url: str, api_key: Optional[str] = None) -> Dict[str, Any]:
    """
    Get the Sonarr session initialisation metadata, including the API key.

    Args:
        host_url (str): Sonarr instance URL.
        api_key (str): Sonarr instance API key, if required. Defaults to `None`.

    Returns:
        Session initialisation metadata
    """

    url = f"{host_url}/initialize.js"
    logger.debug("GET %s", url)
    res = requests.get(
        url,
        headers={"X-Api-Key": api_key} if api_key else None,
        timeout=state.config.buildarr.request_timeout,
    )
    res_match = re.match(INITIALIZE_JS_RES_PATTERN, res.text)
    if not res_match:
        raise RuntimeError(f"No matches for initialize.js parsing: {res.text}")
    res_json = json5.loads(res_match.group(1))
    logger.debug("GET %s -> status_code=%i res=%s", url, res.status_code, repr(res_json))
    return res_json


def api_get(secrets: SonarrSecrets, api_url: str) -> Any:
    """
    Send a `GET` request to a Sonarr instance.

    Args:
        secrets (SonarrSecrets): Sonarr secrets metadata
        api_url (str): Sonarr API command

    Returns:
        Response object
    """

    url = f"{secrets.host_url}/{api_url.lstrip('/')}"
    logger.debug("GET %s", url)
    res = requests.get(
        url,
        headers={"X-Api-Key": secrets.api_key.get_secret_value()},
        timeout=state.config.buildarr.request_timeout,
    )
    res_json = res.json()
    logger.debug("GET %s -> status_code=%i res=%s", url, res.status_code, repr(res_json))
    if res.status_code != HTTPStatus.OK:
        api_error(method="GET", url=url, response=res)
    return res_json


def api_post(secrets: SonarrSecrets, api_url: str, req: Any) -> Any:
    """
    Send a `POST` request to a Sonarr instance.

    Args:
        secrets (SonarrSecrets): Sonarr secrets metadata
        api_url (str): Sonarr API command
        req (Any): Request (JSON-serialisable)

    Returns:
        Response object
    """

    url = f"{secrets.host_url}/{api_url.lstrip('/')}"
    logger.debug("POST %s <- req=%s", url, repr(req))
    headers = {"X-Api-Key": secrets.api_key.get_secret_value()}
    if not state.dry_run:
        res = requests.post(
            url,
            headers=headers,
            json=req,
            timeout=state.config.buildarr.request_timeout,
        )
    else:
        res = _create_dryrun_response("POST", url, content=json.dumps(req))
    res_json = res.json()
    logger.debug("POST %s -> status_code=%i res=%s", url, res.status_code, repr(res_json))
    if res.status_code != HTTPStatus.CREATED:
        api_error(method="POST", url=url, response=res)
    return res_json


def api_put(secrets: SonarrSecrets, api_url: str, req: Any) -> Any:
    """
    Send a `PUT` request to a Sonarr instance.

    Args:
        secrets (SonarrSecrets): Sonarr secrets metadata
        api_url (str): Sonarr API command
        req (Any): Request (JSON-serialisable)

    Returns:
        Response object
    """

    url = f"{secrets.host_url}/{api_url.lstrip('/')}"
    logger.debug("PUT %s <- req=%s", url, repr(req))
    headers = {"X-Api-Key": secrets.api_key.get_secret_value()}
    if not state.dry_run:
        res = requests.put(
            url,
            headers=headers,
            json=req,
            timeout=state.config.buildarr.request_timeout,
        )
    else:
        res = _create_dryrun_response("PUT", url, content=json.dumps(req))
    res_json = res.json()
    logger.debug("PUT %s -> status_code=%i res=%s", url, res.status_code, repr(res_json))
    if res.status_code != HTTPStatus.ACCEPTED:
        api_error(method="PUT", url=url, response=res)
    return res_json


def api_delete(secrets: SonarrSecrets, api_url: str) -> None:
    """
    Send a `DELETE` request to a Sonarr instance.

    Args:
        secrets (SonarrSecrets): Sonarr secrets metadata
        api_url (str): Sonarr API command
    """

    url = f"{secrets.host_url}/{api_url.lstrip('/')}"
    logger.debug("DELETE %s", url)
    headers = {"X-Api-Key": secrets.api_key.get_secret_value()}
    res = (
        requests.delete(
            url,
            headers=headers,
            timeout=state.config.buildarr.request_timeout,
        )
        if not state.dry_run
        else _create_dryrun_response("DELETE", url)
    )
    logger.debug("DELETE %s -> status_code=%i", url, res.status_code)
    if res.status_code != HTTPStatus.OK:
        api_error(method="DELETE", url=url, response=res, parse_response=False)


def api_error(
    method: str,
    url: str,
    response: requests.Response,
    parse_response: bool = True,
) -> None:
    """
    Process an error response from the Sonarr API.

    Args:
        method (str): HTTP method.
        url (str): API command URL.
        response (requests.Response): Response metadata.
        parse_response (bool, optional): Parse response error JSON. Defaults to True.

    Raises:
        Sonarr API exception
    """

    error_message = (
        f"Unexpected response with status code {response.status_code} from from '{method} {url}':"
    )
    if parse_response:
        res_json = response.json()
        try:
            error_message += f" {_api_error(res_json)}"
        except TypeError:
            for error in res_json:
                error_message += f"\n{_api_error(error)}"
        except KeyError:
            error_message += f" {res_json}"
    raise SonarrAPIError(error_message, response=response)


def _api_error(res_json: Any) -> str:
    """
    Generate an error message from a response object.

    Args:
        res_json (Any): Response object

    Returns:
        String containing one or more error messages
    """

    try:
        try:
            error_message = f"{res_json['propertyName']}: {res_json['errorMessage']}"
            try:
                error_message += f" (attempted value: {res_json['attemptedValue']})"
            except KeyError:
                pass
            return error_message
        except KeyError:
            pass
        try:
            return f"{res_json['message']}\n{res_json['description']}"
        except KeyError:
            pass
        return res_json["message"]
    except KeyError:
        return f"(Unsupported error JSON format) {res_json}"


def _create_dryrun_response(
    method: str,
    url: str,
    headers: Optional[Mapping[str, str]] = None,
    status_code: Optional[int] = None,
    content_type: str = "application/json",
    charset: str = "utf-8",
    content: str = "{}",
) -> requests.Response:
    """
    A utility function for generating `requests.Response` objects in dry-run mode.

    Args:
        method (str): HTTP method of the response to simulate.
        url (str): URL of the request.
        status_code (Optional[int], optional): Status code for the response. Default: auto-detect
        content_type (str, optional): MIME type of response content. Default: `application/json`
        charset (str, optional): Encoding of response content. Default: `utf-8`
        content (str, optional): Response content. Default: `{}`

    Raises:
        ValueError: When an unsupported HTTP method is used

    Returns:
        Generated `requests.Response` object
    """

    method = method.upper()

    response = requests.Response()
    response.url = url
    response.headers["Vary"] = "Accept"
    response.headers["Cache-Control"] = "no-cache, no-store, must-revalidate, max-age=0"
    response.headers["Pragma"] = "no-cache"
    response.headers["Expires"] = "0"
    response.headers["Access-Control-Allow-Origin"] = "*"
    response.headers["Content-Type"] = f"{content_type}; charset={charset}"
    response.headers["Server"] = "Mono-HTTPAPI/1.0"
    response.headers["Date"] = datetime.now(tz=timezone.utc).strftime("%a, %d %b %Y %H:%M:%S %Z")
    response.headers["Transfer-Encoding"] = "chunked"
    if headers:
        response.headers.update(headers)
    if status_code is not None:
        response.status_code = status_code
    elif method == "POST":
        response.status_code = int(HTTPStatus.CREATED)
    elif method == "PUT":
        response.status_code = int(HTTPStatus.ACCEPTED)
    elif method == "DELETE":
        response.status_code = int(HTTPStatus.OK)
    else:
        raise ValueError(
            f"Unsupported HTTP method for creating dry-run response: {str(method)}",
        )
    response.encoding = charset
    if content is not None:
        response._content = content.encode("UTF-8")

    return response
